# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
import copy
import itertools
import json
import os.path

from pyomo.common.dependencies import (
    numpy as np,
    numpy_available,
    pandas as pd,
    pandas_available,
    scipy_available,
)

from pyomo.common.fileutils import this_file_dir
import pyomo.common.unittest as unittest

if not (numpy_available and scipy_available):
    raise unittest.SkipTest("Pyomo.DoE needs scipy and numpy to run tests")

if scipy_available:
    from pyomo.contrib.doe import DesignOfExperiments, FIMExternalGreyBox
    from pyomo.contrib.doe.examples.reactor_example import (
        ReactorExperiment as FullReactorExperiment,
    )
    from pyomo.contrib.parmest.examples.rooney_biegler.rooney_biegler import (
        RooneyBieglerExperiment,
    )

import pyomo.environ as pyo

from pyomo.opt import SolverFactory

ipopt_available = SolverFactory("ipopt").available()
cyipopt_available = SolverFactory("cyipopt").available()

currdir = this_file_dir()
file_path = os.path.join(currdir, "..", "examples", "result.json")

with open(file_path) as f:
    data_ex = json.load(f)

data_ex["control_points"] = {float(k): v for k, v in data_ex["control_points"].items()}

_FD_EPSILON_FIRST = 1e-6  # Epsilon for numerical comparison of derivatives
_FD_EPSILON_SECOND = 1e-4  # Epsilon for numerical comparison of derivatives

if numpy_available:
    # Randomly generated P.S.D. matrix
    # Matrix is 4x4 to match example
    # number of parameters.
    testing_matrix = np.array(
        [
            [5.13730123, 1.08084953, 1.6466824, 1.09943223],
            [1.08084953, 1.57183404, 1.50704403, 1.4969689],
            [1.6466824, 1.50704403, 2.54754738, 1.39902838],
            [1.09943223, 1.4969689, 1.39902838, 1.57406692],
        ]
    )

    masking_matrix = np.triu(np.ones_like(testing_matrix))


def get_numerical_derivative(grey_box_object=None):
    # Internal import to avoid circular imports
    from pyomo.contrib.doe import ObjectiveLib

    # Grab current FIM value
    current_FIM = grey_box_object._get_FIM()
    dim = current_FIM.shape[0]
    unperturbed_value = 0

    # Find the initial value of the function
    if grey_box_object.objective_option == ObjectiveLib.trace:
        unperturbed_value = np.trace(np.linalg.inv(current_FIM))
    elif grey_box_object.objective_option == ObjectiveLib.determinant:
        unperturbed_value = np.log(np.linalg.det(current_FIM))
    elif grey_box_object.objective_option == ObjectiveLib.minimum_eigenvalue:
        vals_init, vecs_init = np.linalg.eig(current_FIM)
        unperturbed_value = np.min(vals_init)
    elif grey_box_object.objective_option == ObjectiveLib.condition_number:
        vals_init, vecs_init = np.linalg.eig(current_FIM)
        unperturbed_value = np.log(np.abs(np.max(vals_init) / np.min(vals_init)))

    # Calculate the numerical derivative, using forward difference
    numerical_derivative = np.zeros_like(current_FIM)

    # perturb each direction
    for i in range(dim):
        for j in range(dim):
            FIM_perturbed = copy.deepcopy(current_FIM)
            FIM_perturbed[i, j] += _FD_EPSILON_FIRST

            new_value_ij = 0
            # Test which method is being used:
            if grey_box_object.objective_option == ObjectiveLib.trace:
                new_value_ij = np.trace(np.linalg.inv(FIM_perturbed))
            elif grey_box_object.objective_option == ObjectiveLib.determinant:
                new_value_ij = np.log(np.linalg.det(FIM_perturbed))
            elif grey_box_object.objective_option == ObjectiveLib.minimum_eigenvalue:
                vals, vecs = np.linalg.eig(FIM_perturbed)
                new_value_ij = np.min(vals)
            elif grey_box_object.objective_option == ObjectiveLib.condition_number:
                vals, vecs = np.linalg.eig(FIM_perturbed)
                new_value_ij = np.log(np.abs(np.max(vals) / np.min(vals)))

            # Calculate the derivative value from forward difference
            diff = (new_value_ij - unperturbed_value) / _FD_EPSILON_FIRST

            numerical_derivative[i, j] = diff

    return numerical_derivative


def get_numerical_second_derivative(grey_box_object=None, return_reduced=True):
    # Internal import to avoid circular imports
    from pyomo.contrib.doe import ObjectiveLib

    # Grab current FIM value
    current_FIM = grey_box_object._get_FIM()
    dim = current_FIM.shape[0]

    # Calculate the numerical derivative,
    # using second order formula
    numerical_derivative = np.zeros([dim, dim, dim, dim])

    # perturb each direction
    for i in range(dim):
        for j in range(dim):
            for k in range(dim):
                for l in range(dim):
                    FIM_perturbed_1 = copy.deepcopy(current_FIM)
                    FIM_perturbed_2 = copy.deepcopy(current_FIM)
                    FIM_perturbed_3 = copy.deepcopy(current_FIM)
                    FIM_perturbed_4 = copy.deepcopy(current_FIM)

                    # Need 4 perturbations to cover the
                    # formula H[i, j] = [(FIM + eps (both))
                    # + (FIM +/- eps one each)
                    # + (FIM -/+ eps one each)
                    # + (FIM - eps (both))] / (4*eps**2)
                    FIM_perturbed_1[i, j] += _FD_EPSILON_SECOND
                    FIM_perturbed_1[k, l] += _FD_EPSILON_SECOND

                    FIM_perturbed_2[i, j] += _FD_EPSILON_SECOND
                    FIM_perturbed_2[k, l] += -_FD_EPSILON_SECOND

                    FIM_perturbed_3[i, j] += -_FD_EPSILON_SECOND
                    FIM_perturbed_3[k, l] += _FD_EPSILON_SECOND

                    FIM_perturbed_4[i, j] += -_FD_EPSILON_SECOND
                    FIM_perturbed_4[k, l] += -_FD_EPSILON_SECOND

                    new_values = np.array([0.0, 0.0, 0.0, 0.0])
                    # Test which method is being used:
                    if grey_box_object.objective_option == ObjectiveLib.trace:
                        new_values[0] = np.trace(np.linalg.inv(FIM_perturbed_1))
                        new_values[1] = np.trace(np.linalg.inv(FIM_perturbed_2))
                        new_values[2] = np.trace(np.linalg.inv(FIM_perturbed_3))
                        new_values[3] = np.trace(np.linalg.inv(FIM_perturbed_4))
                    elif grey_box_object.objective_option == ObjectiveLib.determinant:
                        new_values[0] = np.log(np.linalg.det(FIM_perturbed_1))
                        new_values[1] = np.log(np.linalg.det(FIM_perturbed_2))
                        new_values[2] = np.log(np.linalg.det(FIM_perturbed_3))
                        new_values[3] = np.log(np.linalg.det(FIM_perturbed_4))
                    elif (
                        grey_box_object.objective_option
                        == ObjectiveLib.minimum_eigenvalue
                    ):
                        vals, vecs = np.linalg.eig(FIM_perturbed_1)
                        new_values[0] = np.min(vals)
                        vals, vecs = np.linalg.eig(FIM_perturbed_2)
                        new_values[1] = np.min(vals)
                        vals, vecs = np.linalg.eig(FIM_perturbed_3)
                        new_values[2] = np.min(vals)
                        vals, vecs = np.linalg.eig(FIM_perturbed_4)
                        new_values[3] = np.min(vals)
                    elif (
                        grey_box_object.objective_option
                        == ObjectiveLib.condition_number
                    ):
                        vals, vecs = np.linalg.eig(FIM_perturbed_1)
                        new_values[0] = np.log(np.abs(np.max(vals) / np.min(vals)))
                        vals, vecs = np.linalg.eig(FIM_perturbed_2)
                        new_values[1] = np.log(np.abs(np.max(vals) / np.min(vals)))
                        vals, vecs = np.linalg.eig(FIM_perturbed_3)
                        new_values[2] = np.log(np.abs(np.max(vals) / np.min(vals)))
                        vals, vecs = np.linalg.eig(FIM_perturbed_4)
                        new_values[3] = np.log(np.abs(np.max(vals) / np.min(vals)))

                    # Calculate the derivative value from second order difference formula
                    diff = (
                        new_values[0] - new_values[1] - new_values[2] + new_values[3]
                    ) / (4 * _FD_EPSILON_SECOND**2)

                    numerical_derivative[i, j, k, l] = diff

    if return_reduced:
        # This considers a 4-parameter system
        # which is what these tests are based
        # upon. This can be generalized but
        # requires checking the parameter length.
        #
        # Make ordered quads with no repeats
        # of the ordered pairs
        ordered_pairs = itertools.product(range(4), repeat=2)
        ordered_pairs_list = list(itertools.combinations_with_replacement(range(4), 2))
        ordered_quads = itertools.combinations_with_replacement(ordered_pairs, 2)

        numerical_derivative_reduced = np.zeros((10, 10))

        for curr_quad in ordered_quads:
            d1, d2 = curr_quad
            i, j, k, l = d1[0], d1[1], d2[0], d2[1]

            reordered_ijkl = grey_box_object._reorder_pairs(i, j, k, l)
            row = ordered_pairs_list.index((reordered_ijkl[2], reordered_ijkl[3]))
            col = ordered_pairs_list.index((reordered_ijkl[0], reordered_ijkl[1]))

            numerical_derivative_reduced[row, col] += numerical_derivative[i, j, k, l]

            # Duplicate check and addition
            if ((i != j) and (k != l)) and ((i == l) and (j == k)):
                numerical_derivative_reduced[row, col] += numerical_derivative[
                    i, j, k, l
                ]

        numerical_derivative_reduced += (
            numerical_derivative_reduced.transpose()
            - np.diag(np.diag(numerical_derivative_reduced))
        )
        return numerical_derivative_reduced

    # Otherwise return numerical derivative as normal
    return numerical_derivative


def get_standard_args(experiment, fd_method, obj_used):
    args = {}
    args['experiment'] = experiment
    args['fd_formula'] = fd_method
    args['step'] = 1e-3
    args['objective_option'] = obj_used
    args['scale_constant_value'] = 1
    args['scale_nominal_param_value'] = True
    args['prior_FIM'] = None
    args['jac_initial'] = None
    args['fim_initial'] = None
    args['L_diagonal_lower_bound'] = 1e-7
    solver = SolverFactory("ipopt")
    args['solver'] = solver
    # Make solver object with
    # good linear subroutines
    solver = SolverFactory("ipopt")
    solver.options["linear_solver"] = "ma57"
    solver.options["halt_on_ampl_error"] = "yes"
    solver.options["max_iter"] = 3000
    args['solver'] = solver
    # Make greybox solver object with
    # good linear subroutines
    grey_box_solver = SolverFactory("cyipopt")
    grey_box_solver.config.options["linear_solver"] = "ma57"
    grey_box_solver.config.options['tol'] = 1e-4
    grey_box_solver.config.options['mu_strategy'] = "monotone"
    args['grey_box_solver'] = grey_box_solver
    args['tee'] = False
    args['grey_box_tee'] = True
    args['get_labeled_model_args'] = None
    args['_Cholesky_option'] = True
    args['_only_compute_fim_lower'] = True
    return args


def make_greybox_and_doe_objects(objective_option):
    fd_method = "central"
    obj_used = objective_option

    experiment = FullReactorExperiment(data_ex, 10, 3)

    DoE_args = get_standard_args(experiment, fd_method, obj_used)
    DoE_args["use_grey_box_objective"] = True
    DoE_args["prior_FIM"] = testing_matrix

    doe_obj = DesignOfExperiments(**DoE_args)
    doe_obj.create_doe_model()

    grey_box_object = FIMExternalGreyBox(
        doe_object=doe_obj, objective_option=doe_obj.objective_option
    )

    return doe_obj, grey_box_object


def make_greybox_and_doe_objects_rooney_biegler(objective_option):
    fd_method = "central"
    obj_used = objective_option

    data = pd.DataFrame(data=[[2, 10.3]], columns=['hour', 'y'])
    theta = {'asymptote': 19.143, 'rate_constant': 0.5311}

    experiment = RooneyBieglerExperiment(
        data=data.loc[0, :], theta=theta, measure_error=1
    )

    DoE_args = get_standard_args(experiment, fd_method, obj_used)
    DoE_args["use_grey_box_objective"] = True

    # Make a custom grey box solver
    grey_box_solver = SolverFactory("cyipopt")
    grey_box_solver.config.options["linear_solver"] = "ma57"
    grey_box_solver.config.options['tol'] = 1e-4
    grey_box_solver.config.options['mu_strategy'] = "monotone"

    # Add the grey box solver to DoE_args
    DoE_args["grey_box_solver"] = grey_box_solver

    data = pd.DataFrame(data=[[1, 8.3], [7, 19.8]], columns=['hour', 'y'])
    theta = {'asymptote': 19.143, 'rate_constant': 0.5311}
    FIM_prior = np.zeros((2, 2))
    # Calculate prior using existing experiments
    for i in range(len(data)):
        prev_experiment = RooneyBieglerExperiment(
            data=data.loc[i, :], theta=theta, measure_error=1
        )
        doe_obj = DesignOfExperiments(
            **get_standard_args(prev_experiment, fd_method, obj_used)
        )

        FIM_prior += doe_obj.compute_FIM(method='sequential')
    DoE_args["prior_FIM"] = FIM_prior

    doe_obj = DesignOfExperiments(**DoE_args)
    doe_obj.create_doe_model()

    grey_box_object = FIMExternalGreyBox(
        doe_object=doe_obj, objective_option=doe_obj.objective_option
    )

    return doe_obj, grey_box_object


# Test whether or not cyipopt
# is appropriately calling the
# linear solvers.
bad_message = "Invalid option encountered."
cyipopt_call_working = True
if (
    numpy_available
    and scipy_available
    and ipopt_available
    and cyipopt_available
    and pandas_available
):
    try:
        objective_option = "determinant"
        doe_object, _ = make_greybox_and_doe_objects_rooney_biegler(
            objective_option=objective_option
        )

        # Use the grey box objective
        doe_object.use_grey_box = True

        # Run for 1 iteration to see if
        # cyipopt was called.
        doe_object.grey_box_solver.config.options["max_iter"] = 1

        doe_object.run_doe()

        cyipopt_call_working = not (
            bad_message in doe_object.results["Termination Message"]
        )
    except:
        cyipopt_call_working = False


@unittest.skipIf(not ipopt_available, "The 'ipopt' command is not available")
@unittest.skipIf(not numpy_available, "Numpy is not available")
@unittest.skipIf(not scipy_available, "scipy is not available")
@unittest.skipIf(not cyipopt_available, "'cyipopt' is not available")
class TestFIMExternalGreyBox(unittest.TestCase):
    # Test that we can properly
    # set the inputs for the
    # Grey Box object
    def test_set_inputs(self):
        objective_option = "trace"
        doe_obj, grey_box_object = make_greybox_and_doe_objects(
            objective_option=objective_option
        )

        # Set input values to the random testing matrix
        grey_box_object.set_input_values(testing_matrix[masking_matrix > 0])

        # Grab the values from get_FIM
        grey_box_FIM = grey_box_object._get_FIM()

        self.assertTrue(np.all(np.isclose(grey_box_FIM, testing_matrix)))

    # Testing that getting the
    # input names works properly
    def test_input_names(self):
        objective_option = "trace"
        doe_obj, grey_box_object = make_greybox_and_doe_objects(
            objective_option=objective_option
        )

        # Hard-coded names of the inputs, in the order we expect
        input_names = [
            ('A1', 'A1'),
            ('A1', 'A2'),
            ('A1', 'E1'),
            ('A1', 'E2'),
            ('A2', 'A2'),
            ('A2', 'E1'),
            ('A2', 'E2'),
            ('E1', 'E1'),
            ('E1', 'E2'),
            ('E2', 'E2'),
        ]

        # Grabbing input names from grey box object
        input_names_gb = grey_box_object.input_names()

        self.assertListEqual(input_names, input_names_gb)

    # Testing that getting the
    # output names works properly
    def test_output_names(self):
        # Need to test for each objective type
        # A-opt
        objective_option = "trace"
        doe_obj_A, grey_box_object_A = make_greybox_and_doe_objects(
            objective_option=objective_option
        )

        # D-opt
        objective_option = "determinant"
        doe_obj_D, grey_box_object_D = make_greybox_and_doe_objects(
            objective_option=objective_option
        )

        # E-opt
        objective_option = "minimum_eigenvalue"
        doe_obj_E, grey_box_object_E = make_greybox_and_doe_objects(
            objective_option=objective_option
        )

        # ME-opt
        objective_option = "condition_number"
        doe_obj_ME, grey_box_object_ME = make_greybox_and_doe_objects(
            objective_option=objective_option
        )

        # Hard-coded names of the outputs
        # There is one element per
        # objective type
        output_names = ['A-opt', 'log-D-opt', 'E-opt', 'ME-opt']

        # Grabbing input names from grey box object
        output_names_gb = []
        output_names_gb.extend(grey_box_object_A.output_names())
        output_names_gb.extend(grey_box_object_D.output_names())
        output_names_gb.extend(grey_box_object_E.output_names())
        output_names_gb.extend(grey_box_object_ME.output_names())

        self.assertListEqual(output_names, output_names_gb)

    # Testing output computation
    def test_outputs_A_opt(self):
        objective_option = "trace"
        doe_obj, grey_box_object = make_greybox_and_doe_objects(
            objective_option=objective_option
        )

        # Set input values to the random testing matrix
        grey_box_object.set_input_values(testing_matrix[masking_matrix > 0])

        grey_box_A_opt = grey_box_object.evaluate_outputs()

        A_opt = np.trace(np.linalg.inv(testing_matrix))

        self.assertTrue(np.isclose(grey_box_A_opt, A_opt))

    def test_outputs_D_opt(self):
        objective_option = "determinant"
        doe_obj, grey_box_object = make_greybox_and_doe_objects(
            objective_option=objective_option
        )

        # Set input values to the random testing matrix
        grey_box_object.set_input_values(testing_matrix[masking_matrix > 0])

        grey_box_D_opt = grey_box_object.evaluate_outputs()

        D_opt = np.log(np.linalg.det(testing_matrix))

        self.assertTrue(np.isclose(grey_box_D_opt, D_opt))

    def test_outputs_E_opt(self):
        objective_option = "minimum_eigenvalue"
        doe_obj, grey_box_object = make_greybox_and_doe_objects(
            objective_option=objective_option
        )

        # Set input values to the random testing matrix
        grey_box_object.set_input_values(testing_matrix[masking_matrix > 0])

        grey_box_E_opt = grey_box_object.evaluate_outputs()

        vals, vecs = np.linalg.eig(testing_matrix)
        E_opt = np.min(vals)

        self.assertTrue(np.isclose(grey_box_E_opt, E_opt))

    def test_outputs_ME_opt(self):
        objective_option = "condition_number"
        doe_obj, grey_box_object = make_greybox_and_doe_objects(
            objective_option=objective_option
        )

        # Set input values to the random testing matrix
        grey_box_object.set_input_values(testing_matrix[masking_matrix > 0])

        grey_box_ME_opt = grey_box_object.evaluate_outputs()

        vals, vecs = np.linalg.eig(testing_matrix)
        ME_opt = np.log(np.abs(np.max(vals) / np.min(vals)))

        self.assertTrue(np.isclose(grey_box_ME_opt, ME_opt))

    # Testing Jacobian computation
    def test_jacobian_A_opt(self):
        objective_option = "trace"
        doe_obj, grey_box_object = make_greybox_and_doe_objects(
            objective_option=objective_option
        )

        # Set input values to the random testing matrix
        grey_box_object.set_input_values(testing_matrix[masking_matrix > 0])

        # Grab the Jacobian values
        utri_vals_jac = grey_box_object.evaluate_jacobian_outputs().toarray()

        # Recover the Jacobian in Matrix Form
        jac = np.zeros_like(grey_box_object._get_FIM())
        jac[np.triu_indices_from(jac)] = utri_vals_jac
        jac += jac.transpose()
        jac = jac / 2

        # Get numerical derivative matrix
        jac_FD = get_numerical_derivative(grey_box_object)

        # assert that each component is close
        self.assertTrue(np.all(np.isclose(jac, jac_FD, rtol=1e-4, atol=1e-4)))

    def test_jacobian_D_opt(self):
        objective_option = "determinant"
        doe_obj, grey_box_object = make_greybox_and_doe_objects(
            objective_option=objective_option
        )

        # Set input values to the random testing matrix
        grey_box_object.set_input_values(testing_matrix[masking_matrix > 0])

        # Grab the Jacobian values
        utri_vals_jac = grey_box_object.evaluate_jacobian_outputs().toarray()

        # Recover the Jacobian in Matrix Form
        jac = np.zeros_like(grey_box_object._get_FIM())
        jac[np.triu_indices_from(jac)] = utri_vals_jac
        jac += jac.transpose()
        jac = jac / 2

        # Get numerical derivative matrix
        jac_FD = get_numerical_derivative(grey_box_object)

        # assert that each component is close
        self.assertTrue(np.all(np.isclose(jac, jac_FD, rtol=1e-4, atol=1e-4)))

    def test_jacobian_E_opt(self):
        objective_option = "minimum_eigenvalue"
        doe_obj, grey_box_object = make_greybox_and_doe_objects(
            objective_option=objective_option
        )

        # Set input values to the random testing matrix
        grey_box_object.set_input_values(testing_matrix[masking_matrix > 0])

        # Grab the Jacobian values
        utri_vals_jac = grey_box_object.evaluate_jacobian_outputs().toarray()

        # Recover the Jacobian in Matrix Form
        jac = np.zeros_like(grey_box_object._get_FIM())
        jac[np.triu_indices_from(jac)] = utri_vals_jac
        jac += jac.transpose()
        jac = jac / 2

        # Get numerical derivative matrix
        jac_FD = get_numerical_derivative(grey_box_object)

        # assert that each component is close
        self.assertTrue(np.all(np.isclose(jac, jac_FD, rtol=1e-4, atol=1e-4)))

    def test_jacobian_ME_opt(self):
        objective_option = "condition_number"
        doe_obj, grey_box_object = make_greybox_and_doe_objects(
            objective_option=objective_option
        )

        # Set input values to the random testing matrix
        grey_box_object.set_input_values(testing_matrix[masking_matrix > 0])

        # Grab the Jacobian values
        utri_vals_jac = grey_box_object.evaluate_jacobian_outputs().toarray()

        # Recover the Jacobian in Matrix Form
        jac = np.zeros_like(grey_box_object._get_FIM())
        jac[np.triu_indices_from(jac)] = utri_vals_jac
        jac += jac.transpose()
        jac = jac / 2

        # Get numerical derivative matrix
        jac_FD = get_numerical_derivative(grey_box_object)

        # assert that each component is close
        self.assertTrue(np.all(np.isclose(jac, jac_FD, rtol=1e-4, atol=1e-4)))

    # Testing Hessian Computation
    def test_hessian_A_opt(self):
        objective_option = "trace"
        doe_obj, grey_box_object = make_greybox_and_doe_objects(
            objective_option=objective_option
        )

        # Set input values to the random testing matrix
        grey_box_object.set_input_values(testing_matrix[masking_matrix > 0])

        # Grab the Hessian values
        hess_vals_from_gb = grey_box_object.evaluate_hessian_outputs().toarray()

        # Recover the Hessian in Matrix Form
        hess_gb = hess_vals_from_gb
        hess_gb += hess_gb.transpose() - np.diag(np.diag(hess_gb))

        # Get numerical derivative matrix
        hess_FD = get_numerical_second_derivative(grey_box_object)

        # assert that each component is close
        self.assertTrue(np.all(np.isclose(hess_gb, hess_FD, rtol=1e-4, atol=1e-4)))

    def test_hessian_D_opt(self):
        objective_option = "determinant"
        doe_obj, grey_box_object = make_greybox_and_doe_objects(
            objective_option=objective_option
        )

        # Set input values to the random testing matrix
        grey_box_object.set_input_values(testing_matrix[masking_matrix > 0])

        # Grab the Hessian values
        hess_vals_from_gb = grey_box_object.evaluate_hessian_outputs().toarray()

        # Recover the Hessian in Matrix Form
        hess_gb = hess_vals_from_gb
        hess_gb += hess_gb.transpose() - np.diag(np.diag(hess_gb))

        # Get numerical derivative matrix
        hess_FD = get_numerical_second_derivative(grey_box_object)

        # assert that each component is close
        self.assertTrue(np.all(np.isclose(hess_gb, hess_FD, rtol=1e-4, atol=1e-4)))

    def test_hessian_E_opt(self):
        objective_option = "minimum_eigenvalue"
        doe_obj, grey_box_object = make_greybox_and_doe_objects(
            objective_option=objective_option
        )

        # Set input values to the random testing matrix
        grey_box_object.set_input_values(testing_matrix[masking_matrix > 0])

        # Grab the Hessian values
        hess_vals_from_gb = grey_box_object.evaluate_hessian_outputs().toarray()

        # Recover the Hessian in Matrix Form
        hess_gb = hess_vals_from_gb
        hess_gb += hess_gb.transpose() - np.diag(np.diag(hess_gb))

        # Get numerical derivative matrix
        hess_FD = get_numerical_second_derivative(grey_box_object)

        # assert that each component is close
        self.assertTrue(np.all(np.isclose(hess_gb, hess_FD, rtol=1e-4, atol=1e-4)))

    def test_hessian_ME_opt(self):
        objective_option = "condition_number"
        doe_obj, grey_box_object = make_greybox_and_doe_objects(
            objective_option=objective_option
        )

        # Set input values to the random testing matrix
        grey_box_object.set_input_values(testing_matrix[masking_matrix > 0])

        # Grab the Hessian values
        hess_vals_from_gb = grey_box_object.evaluate_hessian_outputs().toarray()

        # Recover the Hessian in Matrix Form
        hess_gb = hess_vals_from_gb
        hess_gb += hess_gb.transpose() - np.diag(np.diag(hess_gb))

        # Get numerical derivative matrix
        hess_FD = get_numerical_second_derivative(grey_box_object)

        # assert that each component is close
        self.assertTrue(np.all(np.isclose(hess_gb, hess_FD, rtol=1e-4, atol=1e-4)))

    def test_equality_constraint_names(self):
        objective_option = "condition_number"
        doe_obj, grey_box_object = make_greybox_and_doe_objects(
            objective_option=objective_option
        )

        # Grab equality constraint names
        eq_con_names_gb = grey_box_object.equality_constraint_names()

        # Equality constraint names should be an
        # empty list.
        self.assertListEqual(eq_con_names_gb, [])

    def test_evaluate_equality_constraints(self):
        objective_option = "condition_number"
        doe_obj, grey_box_object = make_greybox_and_doe_objects(
            objective_option=objective_option
        )

        # Grab equality constraint names
        eq_con_vals_gb = grey_box_object.evaluate_equality_constraints()

        # Equality constraint values should be `None`
        # There are no equality constraints.
        self.assertIsNone(eq_con_vals_gb)

    def test_evaluate_jacobian_equality_constraints(self):
        objective_option = "condition_number"
        doe_obj, grey_box_object = make_greybox_and_doe_objects(
            objective_option=objective_option
        )

        # Grab equality constraint names
        jac_eq_con_vals_gb = grey_box_object.evaluate_jacobian_equality_constraints()

        # Jacobian of equality constraints
        # should be `None` as there are no
        # equality constraints
        self.assertIsNone(jac_eq_con_vals_gb)

    def test_evaluate_hessian_equality_constraints(self):
        objective_option = "condition_number"
        doe_obj, grey_box_object = make_greybox_and_doe_objects(
            objective_option=objective_option
        )

        # Grab equality constraint names
        hess_eq_con_vals_gb = grey_box_object.evaluate_hessian_equality_constraints()

        # Jacobian of equality constraints
        # should be `None` as there are no
        # equality constraints
        self.assertIsNone(hess_eq_con_vals_gb)

    # The following few tests will test whether
    # the DoE problem with grey box is built
    # properly.
    def test_A_opt_greybox_build(self):
        objective_option = "trace"
        doe_obj, _ = make_greybox_and_doe_objects(objective_option=objective_option)

        # Build the greybox objective block
        # on the DoE object
        doe_obj.create_grey_box_objective_function()

        # Check to see if each component exists
        all_exist = True

        # Check output and value
        # FIM Initial will be the prior FIM
        # added with the identity matrix.
        A_opt_val = np.trace(np.linalg.inv(testing_matrix + np.eye(4)))

        try:
            A_opt_val_gb = doe_obj.model.obj_cons.egb_fim_block.outputs["A-opt"].value
        except:
            A_opt_val_gb = -10.0  # Trace should never be negative
            all_exist = False

        # Intermediate check for output existence
        self.assertTrue(all_exist)
        self.assertAlmostEqual(A_opt_val, A_opt_val_gb)

        # Check inputs and values
        try:
            input_values = []
            for i in _.input_names():
                input_values.append(doe_obj.model.obj_cons.egb_fim_block.inputs[i]())
        except:
            input_values = np.zeros_like(testing_matrix)
            all_exist = False

        # Final check on existence of inputs
        self.assertTrue(all_exist)
        # Rebuild the current FIM from the input
        # values taken from the egb_fim_block
        current_FIM = np.zeros_like(testing_matrix)
        current_FIM[np.triu_indices_from(current_FIM)] = input_values
        current_FIM += current_FIM.transpose() - np.diag(np.diag(current_FIM))

        self.assertTrue(np.all(np.isclose(current_FIM, testing_matrix + np.eye(4))))

    def test_D_opt_greybox_build(self):
        objective_option = "determinant"
        doe_obj, _ = make_greybox_and_doe_objects(objective_option=objective_option)

        # Build the greybox objective block
        # on the DoE object
        doe_obj.create_grey_box_objective_function()

        # Check to see if each component exists
        all_exist = True

        # Check output and value
        # FIM Initial will be the prior FIM
        # added with the identity matrix.
        D_opt_val = np.log(np.linalg.det(testing_matrix + np.eye(4)))

        try:
            D_opt_val_gb = doe_obj.model.obj_cons.egb_fim_block.outputs[
                "log-D-opt"
            ].value
        except:
            D_opt_val_gb = -100.0  # Determinant should never be negative beyond -64
            all_exist = False

        # Intermediate check for output existence
        self.assertTrue(all_exist)
        self.assertAlmostEqual(D_opt_val, D_opt_val_gb)

        # Check inputs and values
        try:
            input_values = []
            for i in _.input_names():
                input_values.append(doe_obj.model.obj_cons.egb_fim_block.inputs[i]())
        except:
            input_values = np.zeros_like(testing_matrix)
            all_exist = False

        # Final check on existence of inputs
        self.assertTrue(all_exist)
        # Rebuild the current FIM from the input
        # values taken from the egb_fim_block
        current_FIM = np.zeros_like(testing_matrix)
        current_FIM[np.triu_indices_from(current_FIM)] = input_values
        current_FIM += current_FIM.transpose() - np.diag(np.diag(current_FIM))

        self.assertTrue(np.all(np.isclose(current_FIM, testing_matrix + np.eye(4))))

    def test_E_opt_greybox_build(self):
        objective_option = "minimum_eigenvalue"
        doe_obj, _ = make_greybox_and_doe_objects(objective_option=objective_option)

        # Build the greybox objective block
        # on the DoE object
        doe_obj.create_grey_box_objective_function()

        # Check to see if each component exists
        all_exist = True

        # Check output and value
        # FIM Initial will be the prior FIM
        # added with the identity matrix.
        vals, vecs = np.linalg.eig(testing_matrix + np.eye(4))
        E_opt_val = np.min(vals)

        try:
            E_opt_val_gb = doe_obj.model.obj_cons.egb_fim_block.outputs["E-opt"].value
        except:
            E_opt_val_gb = -10.0  # Determinant should never be negative
            all_exist = False

        # Intermediate check for output existence
        self.assertTrue(all_exist)
        self.assertAlmostEqual(E_opt_val, E_opt_val_gb)

        # Check inputs and values
        try:
            input_values = []
            for i in _.input_names():
                input_values.append(doe_obj.model.obj_cons.egb_fim_block.inputs[i]())
        except:
            input_values = np.zeros_like(testing_matrix)
            all_exist = False

        # Final check on existence of inputs
        self.assertTrue(all_exist)
        # Rebuild the current FIM from the input
        # values taken from the egb_fim_block
        current_FIM = np.zeros_like(testing_matrix)
        current_FIM[np.triu_indices_from(current_FIM)] = input_values
        current_FIM += current_FIM.transpose() - np.diag(np.diag(current_FIM))

        self.assertTrue(np.all(np.isclose(current_FIM, testing_matrix + np.eye(4))))

    def test_ME_opt_greybox_build(self):
        objective_option = "condition_number"
        doe_obj, _ = make_greybox_and_doe_objects(objective_option=objective_option)

        # Build the greybox objective block
        # on the DoE object
        doe_obj.create_grey_box_objective_function()

        # Check to see if each component exists
        all_exist = True

        # Check output and value
        # FIM Initial will be the prior FIM
        # added with the identity matrix.
        vals, vecs = np.linalg.eig(testing_matrix + np.eye(4))
        ME_opt_val = np.log(np.abs(np.max(vals) / np.min(vals)))

        try:
            ME_opt_val_gb = doe_obj.model.obj_cons.egb_fim_block.outputs["ME-opt"].value
        except:
            ME_opt_val_gb = -10.0  # Condition number should not be negative
            all_exist = False

        # Intermediate check for output existence
        self.assertTrue(all_exist)
        self.assertAlmostEqual(ME_opt_val, ME_opt_val_gb)

        # Check inputs and values
        try:
            input_values = []
            for i in _.input_names():
                input_values.append(doe_obj.model.obj_cons.egb_fim_block.inputs[i]())
        except:
            input_values = np.zeros_like(testing_matrix)
            all_exist = False

        # Final check on existence of inputs
        self.assertTrue(all_exist)
        # Rebuild the current FIM from the input
        # values taken from the egb_fim_block
        current_FIM = np.zeros_like(testing_matrix)
        current_FIM[np.triu_indices_from(current_FIM)] = input_values
        current_FIM += current_FIM.transpose() - np.diag(np.diag(current_FIM))

        self.assertTrue(np.all(np.isclose(current_FIM, testing_matrix + np.eye(4))))

    # Testing all the error messages
    def test_constructor_doe_object_error(self):
        with self.assertRaisesRegex(
            ValueError,
            "DoE Object must be provided to build external grey box of the FIM.",
        ):
            grey_box_object = FIMExternalGreyBox(doe_object=None)

    def test_constructor_objective_lib_error(self):
        objective_option = "trace"
        doe_object, grey_box_object = make_greybox_and_doe_objects(
            objective_option=objective_option
        )
        with self.assertRaisesRegex(
            ValueError, "'Bad Objective Option' is not a valid ObjectiveLib"
        ):
            bad_grey_box_object = FIMExternalGreyBox(
                doe_object=doe_object, objective_option="Bad Objective Option"
            )

    def test_output_names_obj_lib_error(self):
        objective_option = "trace"
        doe_object, grey_box_object = make_greybox_and_doe_objects(
            objective_option=objective_option
        )

        grey_box_object.objective_option = "Bad Objective Option"

        with self.assertRaisesRegex(
            ValueError, "'Bad Objective Option' is not a valid ObjectiveLib"
        ):
            grey_box_object.output_names()

    def test_evaluate_outputs_obj_lib_error(self):
        objective_option = "trace"
        doe_object, grey_box_object = make_greybox_and_doe_objects(
            objective_option=objective_option
        )

        grey_box_object.objective_option = "Bad Objective Option"

        with self.assertRaisesRegex(
            ValueError, "'Bad Objective Option' is not a valid ObjectiveLib"
        ):
            grey_box_object.evaluate_outputs()

    def test_evaluate_jacobian_outputs_obj_lib_error(self):
        objective_option = "trace"
        doe_object, grey_box_object = make_greybox_and_doe_objects(
            objective_option=objective_option
        )

        grey_box_object.objective_option = "Bad Objective Option"

        with self.assertRaisesRegex(
            ValueError, "'Bad Objective Option' is not a valid ObjectiveLib"
        ):
            grey_box_object.evaluate_jacobian_outputs()

    def test_evaluate_hessian_outputs_obj_lib_error(self):
        objective_option = "trace"
        doe_object, grey_box_object = make_greybox_and_doe_objects(
            objective_option=objective_option
        )

        grey_box_object.objective_option = "Bad Objective Option"

        with self.assertRaisesRegex(
            ValueError, "'Bad Objective Option' is not a valid ObjectiveLib"
        ):
            grey_box_object.evaluate_hessian_outputs()

    # Test all versions of solving
    # using grey box
    @unittest.skipIf(
        not cyipopt_call_working, "cyipopt is not properly accessing linear solvers"
    )
    def test_solve_D_optimality_log_determinant(self):
        # Two locally optimal design points exist
        # (time, optimal objective value)
        # Here, the objective value is
        # log-10(determinant) of the FIM
        optimal_experimental_designs = [
            np.array([1.782, 4.34]),
            np.array([10.00, 4.35]),
        ]
        objective_option = "determinant"
        doe_object, grey_box_object = make_greybox_and_doe_objects_rooney_biegler(
            objective_option=objective_option
        )

        # Set to use the grey box objective
        doe_object.use_grey_box = True

        # Solve the model
        doe_object.run_doe()

        optimal_time_val = doe_object.results["Experiment Design"][0]
        optimal_obj_val = np.log10(np.exp(pyo.value(doe_object.model.objective)))

        optimal_design_np_array = np.array([optimal_time_val, optimal_obj_val])

        self.assertTrue(
            np.all(
                np.isclose(
                    optimal_design_np_array, optimal_experimental_designs[0], 1e-1
                )
            )
            or np.all(
                np.isclose(
                    optimal_design_np_array, optimal_experimental_designs[1], 1e-1
                )
            )
        )

    @unittest.skipIf(
        not cyipopt_call_working, "cyipopt is not properly accessing linear solvers"
    )
    def test_solve_A_optimality_trace_of_inverse(self):
        # Two locally optimal design points exist
        # (time, optimal objective value)
        # Here, the objective value is
        # trace(inverse(FIM))
        optimal_experimental_designs = [
            np.array([1.33, 0.0277]),
            np.array([9.9, 0.0366]),
        ]
        objective_option = "trace"
        doe_object, grey_box_object = make_greybox_and_doe_objects_rooney_biegler(
            objective_option=objective_option
        )

        # Set to use the grey box objective
        doe_object.use_grey_box = True

        # Solve the model
        doe_object.run_doe()

        optimal_time_val = doe_object.results["Experiment Design"][0]
        optimal_obj_val = doe_object.model.objective()

        optimal_design_np_array = np.array([optimal_time_val, optimal_obj_val])

        self.assertTrue(
            np.all(
                np.isclose(
                    optimal_design_np_array, optimal_experimental_designs[0], 1e-1
                )
            )
            or np.all(
                np.isclose(
                    optimal_design_np_array, optimal_experimental_designs[1], 1e-1
                )
            )
        )

    @unittest.skipIf(
        not cyipopt_call_working, "cyipopt is not properly accessing linear solvers"
    )
    @unittest.skipIf(not pandas_available, "pandas is not available")
    def test_solve_E_optimality_minimum_eigenvalue(self):
        # Two locally optimal design points exist
        # (time, optimal objective value)
        # Here, the objective value is
        # minimum eigenvalue of the FIM
        optimal_experimental_designs = [
            np.array([1.30, 38.70]),
            np.array([10.00, 28.349]),
        ]
        objective_option = "minimum_eigenvalue"
        doe_object, grey_box_object = make_greybox_and_doe_objects_rooney_biegler(
            objective_option=objective_option
        )

        # Set to use the grey box objective
        doe_object.use_grey_box = True

        # Solve the model
        doe_object.run_doe()

        optimal_time_val = doe_object.results["Experiment Design"][0]
        optimal_obj_val = doe_object.model.objective()

        optimal_design_np_array = np.array([optimal_time_val, optimal_obj_val])

        self.assertTrue(
            np.all(
                np.isclose(
                    optimal_design_np_array, optimal_experimental_designs[0], 1e-1
                )
            )
            or np.all(
                np.isclose(
                    optimal_design_np_array, optimal_experimental_designs[1], 1e-1
                )
            )
        )

    @unittest.skipIf(
        not cyipopt_call_working, "cyipopt is not properly accessing linear solvers"
    )
    @unittest.skipIf(not pandas_available, "pandas is not available")
    def test_solve_ME_optimality_condition_number(self):
        # Two locally optimal design points exist
        # (time, optimal objective value)
        # Here, the objective value is
        # condition number of the FIM
        optimal_experimental_designs = [
            np.array([0.943, np.log(13.524)]),
            np.array([10.00, np.log(27.675)]),
        ]
        objective_option = "condition_number"
        doe_object, grey_box_object = make_greybox_and_doe_objects_rooney_biegler(
            objective_option=objective_option
        )

        # Set to use the grey box objective
        doe_object.use_grey_box = True

        # Solve the model
        doe_object.run_doe()

        optimal_time_val = doe_object.results["Experiment Design"][0]
        optimal_obj_val = doe_object.model.objective()

        optimal_design_np_array = np.array([optimal_time_val, optimal_obj_val])

        self.assertTrue(
            np.all(
                np.isclose(
                    optimal_design_np_array, optimal_experimental_designs[0], 1e-1
                )
            )
            or np.all(
                np.isclose(
                    optimal_design_np_array, optimal_experimental_designs[1], 1e-1
                )
            )
        )


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

"""This module provides utilities for deprecating functionality.

.. autosummary::

   deprecated
   deprecation_warning
   relocated_module
   relocated_module_attribute
   RenamedClass
"""

import logging
import functools
import importlib
import inspect
import itertools
import sys
import textwrap
import types
import typing

from pyomo.common.errors import DeveloperError
from pyomo.common.flags import NOTSET, in_testing_environment, building_documentation

_doc_flag = '.. deprecated::'


def _autosummary_doctest_setup():
    """This function gets called to setup the doctest environment before
    running any autosummary doctests in this module.
    """

    # Ensure that all logged messages are sent to stdout
    # (so they show up in the doctest output and can be tested)
    import pyomo.common.log as _log

    _log.pyomo_handler.__class__ = _log.StdoutHandler


def default_deprecation_msg(obj, user_msg, version, remove_in):
    """Generate the default deprecation message.

    See deprecated() function for argument details.
    """
    if user_msg is None:
        if inspect.isclass(obj):
            _obj = ' class'
        elif inspect.ismethod(obj):
            _obj = ' method'
        elif inspect.isfunction(obj) or inspect.isbuiltin(obj):
            _obj = ' function'
        else:
            # either @deprecated() an unknown type or called from
            # deprecation_warning()
            _obj = ''

        _qual = getattr(obj, '__qualname__', '') or ''
        if _qual.endswith('.__init__') or _qual.endswith('.__new__'):
            _obj = f' class ({_qual.rsplit(".", 1)[0]})'
        elif _qual and _obj:
            _obj += f' ({_qual})'

        user_msg = (
            'This%s has been deprecated and may be removed in a '
            'future release.' % (_obj,)
        )
    comment = []
    if version:
        comment.append('deprecated in %s' % (version,))
    if remove_in:
        comment.append('will be removed in (or after) %s' % (remove_in))
    if comment:
        return user_msg + "  (%s)" % (', '.join(comment),)
    else:
        return user_msg


def _deprecation_docstring(obj, msg, version, remove_in):
    # Note that _deprecation_docstring is guaranteed to be called by
    # @deprecated in all situations where we would be creating a
    # meaningful deprecation message (classes, functions, and methods),
    # so this is a convenient place to check that the version is
    # specified.
    if version is None:
        raise DeveloperError("@deprecated(): missing 'version' argument")
    return (
        f'{_doc_flag} {version}\n'
        f'   {default_deprecation_msg(obj, msg, None, remove_in)}\n'
    )


def _wrap_class(cls, msg, logger, version, remove_in):
    _doc = None
    # Note: __new_member__ is where enum.Enum buries the user's original
    # __new__ method
    for field in ('__new__', '__init__', '__new_member__'):
        _funcDoc = getattr(getattr(cls, field, None), '__doc__', '') or ''
        _flagIdx = _funcDoc.find(_doc_flag)
        if _flagIdx >= 0:
            _doc = _funcDoc[_flagIdx:]
            break
    # Note: test 'msg is not None' to revert back to the user-supplied
    # message.  Checking the fields above is still useful as it lets us know
    # if there is already a deprecation message on either new or init.
    if msg is not None or _doc is None:
        _doc = _deprecation_docstring(cls, msg, version, remove_in)
    if cls.__doc__:
        _doc = inspect.cleandoc(cls.__doc__) + '\n\n' + _doc
    cls.__doc__ = 'DEPRECATED.\n\n' + _doc

    if _flagIdx < 0:
        # No deprecation message on __init__ or __new__: go through and
        # find the "most derived" implementation of either __new__ or
        # __init__ and wrap that (breaking ties in favor of __init__)
        field = '__init__'
        for c in reversed(cls.__mro__):
            for f in ('__new__', '__init__'):
                if getattr(c, f, None) is not getattr(cls, f, None):
                    field = f
        setattr(
            cls, field, _wrap_func(getattr(cls, field), msg, logger, version, remove_in)
        )
    return cls


def _wrap_func(func, msg, logger, version, remove_in):
    message = default_deprecation_msg(func, msg, version, remove_in)

    @functools.wraps(
        func, assigned=('__module__', '__name__', '__qualname__', '__annotations__')
    )
    def wrapper(*args, **kwargs):
        cf = _find_calling_frame(1)
        deprecation_warning(message, logger, version='', calling_frame=cf)
        return func(*args, **kwargs)

    wrapper.__doc__ = 'DEPRECATED.\n\n'
    _doc = inspect.cleandoc(func.__doc__ or '')
    if _doc:
        wrapper.__doc__ += _doc + '\n\n'
    wrapper.__doc__ += _deprecation_docstring(func, msg, version, remove_in)
    return wrapper


def _find_calling_frame(module_offset):
    g = [globals()]
    calling_frame = inspect.currentframe().f_back
    while calling_frame is not None:
        if calling_frame.f_globals is g[-1]:
            calling_frame = calling_frame.f_back
        elif len(g) < module_offset:
            g.append(calling_frame.f_globals)
        else:
            break
    return calling_frame


def deprecation_warning(
    msg, logger=None, version=None, remove_in=None, calling_frame=None
):
    """Standardized function for formatting and emitting deprecation warnings.

    This is a standardized routine for formatting deprecation warnings
    so that things look consistent and "nice".

    Args:
        msg (str): the deprecation message to format

        logger (str): the logger to use for emitting the warning
            (default: the calling pyomo package, or "pyomo")

        version (str): [required] the version in which the decorated
            object was deprecated.  General practice is to set version
            to the current development version (from `pyomo --version`)
            during development and update it to the actual release as
            part of the release process.

        remove_in (str): the version in which the decorated object will be
            removed from the code.

        calling_frame (frame): the original frame context that triggered
            the deprecation warning.

    Example
    -------
    >>> from pyomo.common.deprecation import deprecation_warning
    >>> deprecation_warning('This functionality is deprecated.', version='1.2.3')
    WARNING: DEPRECATED: This functionality is deprecated.  (deprecated in 1.2.3) ...

    """
    if version is None:
        raise DeveloperError("deprecation_warning() missing 'version' argument")

    if logger is None:
        if calling_frame is not None:
            cf = calling_frame
        else:
            # The relevant module is the one that holds the
            # function/method that called deprecation_warning
            cf = _find_calling_frame(1)
        if cf is not None:
            logger = cf.f_globals.get('__name__', None)
            if logger is not None and not logger.startswith('pyomo'):
                logger = None
        if logger is None:
            logger = 'pyomo'
    if isinstance(logger, str):
        logger = logging.getLogger(logger)

    msg = textwrap.fill(
        f'DEPRECATED: {default_deprecation_msg(None, msg, version, remove_in)}',
        width=70,
    )
    if calling_frame is None:
        # The useful thing to let the user know is what called the
        # function that generated the deprecation warning.  The current
        # globals() is *this* module.  Walking up the stack to find the
        # frame where the globals() changes tells us the module that is
        # issuing the deprecation warning.  As we assume that *that*
        # module will not trigger its own deprecation warnings, we will
        # walk farther up until the globals() changes again.
        calling_frame = _find_calling_frame(2)
    if calling_frame is not None:
        info = inspect.getframeinfo(calling_frame)
        msg += "\n(called from %s:%s)" % (info.filename.strip(), info.lineno)
        if deprecation_warning.emitted_warnings is not None:
            if msg in deprecation_warning.emitted_warnings:
                return
            deprecation_warning.emitted_warnings.add(msg)

    logger.warning(msg)


# We do not want to cache / suppress repeated warnings when we are
# testing or when we are building the documentation.  Note that doctest
# doesn't set the "in_testing" flag until after pyomo.common is
# imported.
if in_testing_environment() or building_documentation():
    deprecation_warning.emitted_warnings = None
else:
    deprecation_warning.emitted_warnings = set()


def deprecated(msg=None, logger=None, version=None, remove_in=None):
    """Decorator to indicate that a function, method, or class is deprecated.

    This decorator will cause a warning to be logged when the wrapped
    function or method is called, or when the deprecated class is
    constructed.  This decorator also updates the target object's
    docstring to indicate that it is deprecated.

    Args:
        msg (str): a custom deprecation message (default: "This
            {function|class} has been deprecated and may be
            removed in a future release.")

        logger (str): the logger to use for emitting the warning
            (default: the calling pyomo package, or "pyomo")

        version (str): [required] the version in which the decorated
            object was deprecated.  General practice is to set version
            to the current development version (from `pyomo --version`)
            during development and update it to the actual release as
            part of the release process.

        remove_in (str): the version in which the decorated object will be
            removed from the code.

    Example
    -------
    >>> from pyomo.common.deprecation import deprecated
    >>> @deprecated(version='1.2.3')
    ... def sample_function(x):
    ...     return 2*x
    >>> sample_function(5)
    WARNING: DEPRECATED: This function (sample_function) has been deprecated and
        may be removed in a future release.  (deprecated in 1.2.3) ...
    10

    """

    def wrap(obj):
        if inspect.isclass(obj):
            return _wrap_class(obj, msg, logger, version, remove_in)
        else:
            return _wrap_func(obj, msg, logger, version, remove_in)

    return wrap


def _import_object(name, target, version, remove_in, msg):
    modname, targetname = target.rsplit('.', 1)
    _object = getattr(importlib.import_module(modname), targetname)
    if msg is None:
        if inspect.isclass(_object):
            _type = 'class'
        elif inspect.isfunction(_object):
            _type = 'function'
        else:
            _type = 'attribute'
        msg = (
            f"the '{name}' {_type} has been moved to '{target}'."
            "  Please update your import."
        )
    deprecation_warning(msg, version=version, remove_in=remove_in)
    return _object


@deprecated(
    "relocated_module() has been deprecated.  Please use moved_module()",
    version='6.8.1',
)
def relocated_module(new_name, msg=None, logger=None, version=None, remove_in=None):
    """Provide a deprecation path for moved / renamed modules

    Upon import, the old module (that called `relocated_module()`) will
    be replaced in :data:`sys.modules` by an alias that points directly to the
    new module.  As a result, the old module should have only two lines
    of executable Python code (the import of `relocated_module` and the
    call to it).

    Note
    ----
    This method (which was placed in the old module that is
    being removed) is deprecated and should be replaced by calls to
    :func:`moved_module()`, which can be called in any parent scope of
    the removed module and does not require that the old module continue
    to exist in the project.

    Parameters
    ----------
    new_name: str
        The new (fully-qualified) module name

    msg: str
        A custom deprecation message.

    logger: str
        The logger to use for emitting the warning (default: the calling
        pyomo package, or "pyomo")

    version: str [required]
        The version in which the module was renamed or moved.  General
        practice is to set version to the current development version
        (from `pyomo --version`) during development and update it to the
        actual release as part of the release process.

    remove_in: str
        The version in which the module will be removed from the code.

    Example
    -------
    >>> from pyomo.common.deprecation import relocated_module
    >>> relocated_module('pyomo.common.deprecation', version='1.2.3')
    WARNING: DEPRECATED: ... The '...' module has been moved to
        'pyomo.common.deprecation'. Please update your import.
        (deprecated in 1.2.3) ...

    """
    new_module = importlib.import_module(new_name)

    # The relevant module (the one being deprecated) is the one that
    # holds the function/method that called relocated_module().  The
    # relevant calling frame for the deprecation warning is the first
    # frame in the stack that doesn't look like the importer (i.e., the
    # thing that imported the deprecated module).
    cf = _find_calling_frame(1)
    old_name = cf.f_globals.get('__name__', '<stdin>')
    cf = cf.f_back
    if cf is not None:
        importer = cf.f_back.f_globals['__name__'].split('.')[0]
        while cf is not None and cf.f_globals['__name__'].split('.')[0] == importer:
            cf = cf.f_back
    if cf is None:
        cf = _find_calling_frame(1)

    sys.modules[old_name] = new_module
    if msg is None:
        msg = (
            f"The '{old_name}' module has been moved to '{new_name}'. "
            'Please update your import.'
        )
    deprecation_warning(msg, logger, version, remove_in, cf)


def relocated_module_attribute(
    local, target, version, remove_in=None, msg=None, f_globals=None
):
    """Provide a deprecation path for moved / renamed module attributes

    This function declares that a local module attribute has been moved
    to another location.  For Python 3.7+, it leverages a
    module.__getattr__ method to manage the deferred import of the
    object from the new location (on request), as well as emitting the
    deprecation warning.

    Parameters
    ----------
    local: str
        The original (local) name of the relocated attribute

    target: str
        The new absolute import name of the relocated attribute

    version: str
        The Pyomo version when this move was released
        (passed to deprecation_warning)

    remove_in: str
        The Pyomo version when this deprecation path will be removed
        (passed to deprecation_warning)

    msg: str
        If not None, then this specifies a custom deprecation message to
        be emitted when the attribute is accessed from its original
        location.

    """
    if version is None:
        raise DeveloperError("relocated_module_attribute(): missing 'version' argument")
    # Historical note: This method only works for Python >= 3.7.  There
    # were backports to previous Python interpreters, but were removed
    # after SHA 4e04819aaeefc2c08b7718460918885e12343451
    if f_globals is None:
        f_globals = inspect.currentframe().f_back.f_globals
        if f_globals['__name__'].startswith('importlib.'):
            raise DeveloperError(
                "relocated_module_attribute() called from a cythonized "
                "module without passing f_globals"
            )
    _relocated = f_globals.get('__relocated_attrs__', None)
    if _relocated is None:
        f_globals['__relocated_attrs__'] = _relocated = {}
        _mod_getattr = f_globals.get('__getattr__', None)

        def __getattr__(name):
            info = _relocated.get(name, None)
            if info is not None:
                target_obj = _import_object(name, *info)
                f_globals[name] = target_obj
                return target_obj
            elif _mod_getattr is not None:
                return _mod_getattr(name)
            raise AttributeError(
                "module '%s' has no attribute '%s'" % (f_globals['__name__'], name)
            )

        f_globals['__getattr__'] = __getattr__
    _relocated[local] = (target, version, remove_in, msg)


class RenamedClass(type):
    """Metaclass to provide a deprecation path for renamed classes

    This metaclass provides a mechanism for renaming old classes while
    still preserving isinstance / issubclass relationships.

    Examples
    --------
    >>> from pyomo.common.deprecation import RenamedClass
    >>> class NewClass:
    ...     pass
    >>> class OldClass(metaclass=RenamedClass):
    ...     __renamed__new_class__ = NewClass
    ...     __renamed__version__ = '6.0'

    Deriving from the old class generates a warning:

    >>> class DerivedOldClass(OldClass):
    ...     pass
    WARNING: DEPRECATED: Declaring class 'DerivedOldClass' derived from
        'OldClass'. The class 'OldClass' has been renamed to 'NewClass'.
        (deprecated in 6.0) ...

    As does instantiating the old class:

    >>> old = OldClass()
    WARNING: DEPRECATED: Instantiating class 'OldClass'.  The class
        'OldClass' has been renamed to 'NewClass'.  (deprecated in 6.0) ...

    Finally, `isinstance` and `issubclass` still work, for example:

    >>> isinstance(old, NewClass)
    True
    >>> class NewSubclass(NewClass):
    ...     pass
    >>> new = NewSubclass()
    >>> isinstance(new, OldClass)
    WARNING: DEPRECATED: Checking type relative to 'OldClass'.  The class
        'OldClass' has been renamed to 'NewClass'.  (deprecated in 6.0) ...
    True

    """

    def __new__(cls, name, bases, classdict, *args, **kwargs):
        new_class = classdict.get('__renamed__new_class__', None)
        if new_class is not None:

            def __renamed__new__(cls, *args, **kwargs):
                cls.__renamed__warning__("Instantiating class '%s'." % (cls.__name__,))
                return new_class(*args, **kwargs)

            classdict['__new__'] = __renamed__new__

            def __renamed__warning__(msg):
                version = classdict.get('__renamed__version__')
                remove_in = classdict.get('__renamed__remove_in__')
                deprecation_warning(
                    "%s  The class '%s' has been renamed to '%s'."
                    % (msg, name, new_class.__name__),
                    version=version,
                    remove_in=remove_in,
                    calling_frame=_find_calling_frame(1),
                )

            classdict['__renamed__warning__'] = __renamed__warning__

            if not classdict.get('__renamed__version__'):
                raise DeveloperError(
                    "Declaring class '%s' using the RenamedClass metaclass, "
                    "but without specifying the __renamed__version__ class "
                    "attribute" % (name,)
                )

        renamed_bases = []
        for base in bases:
            new_class = getattr(base, '__renamed__new_class__', None)
            if new_class is not None:
                base.__renamed__warning__(
                    "Declaring class '%s' derived from '%s'." % (name, base.__name__)
                )
                base = new_class
                # Flag that this class is derived from a renamed class
                classdict.setdefault('__renamed__new_class__', None)
            # Avoid duplicates (in case someone does a diamond between
            # the renamed class and [a class derived from] the new
            # class)
            if base not in renamed_bases:
                renamed_bases.append(base)

        # Add the new class as a "base class" of the renamed class (this
        # makes issubclass(renamed, new_class) work correctly).  As we
        # still never create an actual instance of renamed, this doesn't
        # affect the API)
        if new_class is not None and new_class not in renamed_bases:
            renamed_bases.append(new_class)

        if new_class is None and '__renamed__new_class__' not in classdict:
            if not any(
                hasattr(mro, '__renamed__new_class__')
                for mro in itertools.chain.from_iterable(
                    base.__mro__ for base in renamed_bases
                )
            ):
                raise TypeError(
                    "Declaring class '%s' using the RenamedClass metaclass, "
                    "but without specifying the __renamed__new_class__ class "
                    "attribute" % (name,)
                )

        return super().__new__(
            cls, name, tuple(renamed_bases), classdict, *args, **kwargs
        )

    def __instancecheck__(cls, instance):
        # Note: the warning is issued by subclasscheck
        return any(
            cls.__subclasscheck__(c) for c in {type(instance), instance.__class__}
        )

    def __subclasscheck__(cls, subclass):
        if hasattr(cls, '__renamed__warning__'):
            cls.__renamed__warning__(
                "Checking type relative to '%s'." % (cls.__name__,)
            )
        if subclass is cls:
            return True
        elif getattr(cls, '__renamed__new_class__') is not None:
            return issubclass(subclass, getattr(cls, '__renamed__new_class__'))
        else:
            return super().__subclasscheck__(subclass)


class MovedModuleLoader:
    """Custom module loader that supports loading modules through alternate names

    This class implements the :class:`importlib.abc.Loader` interface
    (through duck-typing to avoid a surprisingly costly import of
    :mod:`importlib.abc`).  Calls to :meth:`create_module()` and
    :meth:`exec_module()` are delegated to corresponding methods on the
    loader for the new module name.

    """

    def __init__(self, info):
        self._info = info

    def create_module(self, spec) -> types.ModuleType:
        msg = self._info.msg
        if msg is NOTSET:
            msg = (
                f"The '{spec.name}' module has been moved to '{self._info.new_name}'. "
                'Please update your import.'
            )
        if msg is not None:
            deprecation_warning(
                msg, self._info.logger, self._info.version, self._info.remove_in
            )
        if self._info.new_name in sys.modules:
            return sys.modules[self._info.new_name]
        return importlib.import_module(self._info.new_name)

    def exec_module(self, module: types.ModuleType) -> None:
        pass


class MovedModuleFinder:
    """Custom finder that supports loading a module through an alternative name.

    This class implements the :class:`importlib.abc.Finder` interface
    (through duck-typing to avoid a surprisingly costly import of
    :mod:`importlib.abc`).

    Pyomo automatically registers a single instance of this finder with
    :mod:`importlib` by appending it to the end of the
    :data:`sys.meta_path` list when this module is imported.
    Subsequent calls to :func:`moved_module` register the association
    between the old and new module names with the ``mapping`` class
    attribute.

    """

    mapping = {}
    ":class:`dict` that maps (removed) module names to :class:`MovedModuleInfo` objects"

    def find_spec(self, fullname, path, target=None):
        if fullname not in self.mapping:
            return None

        info = MovedModuleFinder.mapping[fullname]
        src_spec = importlib.util.find_spec(info.new_name)
        return importlib.machinery.ModuleSpec(
            name=fullname,
            loader=MovedModuleLoader(info),
            origin=getattr(src_spec, 'origin', None),
        )

    def invalidate_caches(self):
        pass


# Insert the MovedModuleFinder at the end of the sys.meta_path.  This
# way, it has no impact on the performance of importing "normal"
# (present) modules, and instead is called as a "last-chance" finder
# before Python would raise an ImportError
sys.meta_path.append(MovedModuleFinder())


MovedModuleInfo = typing.NamedTuple(
    'MovedModuleInfo',
    [
        ('old_name', str),
        ('new_name', str),
        ('msg', str),
        ('logger', str),
        ('version', str),
        ('remove_in', str),
    ],
)


def moved_module(
    old_name, new_name, msg=NOTSET, logger=None, version=None, remove_in=None
):
    """Provide a deprecation path for moved / renamed modules

    This function hooks into the Python :mod:`importlib` to cause any
    import of the ``old_name`` to instead import and return the module
    from ``new_name``.  The new module is automatically registered with
    :data:`sys.modules` under both the old and new names.

    Because :func:`moved_module()` works through the Python
    :mod:`importlib` system, the old module file can be completely
    deleted (in contrast to the [deprecated] :func:`relocated_module()`
    function).  Calls to :func:`moved_module` should be placed in any
    package above the removed module in the package hierarchy (or in any
    other location that is guaranteeded to be imported / executed before
    any attempts at importng the module through its old name.

    Any import of the module through the old name will emit a
    deprecation warning unless the ``msg`` is ``None`` (see also
    :func:`deprecated`).

    Parameters
    ----------
    old_name: str
        The original (fully-qualified) module name (that has been removed)

    new_name: str
        The new (fully-qualified) module name

    msg: str
        A custom deprecation message.  If None, the deprecation message
        will be suppressed.  If NOTSET (default), a generic deprecation
        message will be logged.

    logger: str
        The logger to use for emitting the warning (default: the calling
        pyomo package, or "pyomo")

    version: str [required]
        The version in which the module was renamed or moved.  General
        practice is to set version to the current development version
        (from `pyomo --version`) during development and update it to the
        actual release as part of the release process.

    remove_in: str
        The version in which the module will be removed from the code.

    Example
    -------
    >>> from pyomo.common.deprecation import moved_module
    >>> moved_module(
    ...     'pyomo.common.old_deprecation',
    ...     'pyomo.common.deprecation',
    ...     version='1.2.3',
    ... )
    >>> import pyomo.common.old_deprecation
    WARNING: DEPRECATED: The 'pyomo.common.old_deprecation' module has
        been moved to 'pyomo.common.deprecation'. Please update your import.
        (deprecated in 1.2.3) ...
    >>> import pyomo.common.deprecation
    >>> pyomo.common.old_deprecation is pyomo.common.deprecation
    True

    """
    if old_name in MovedModuleFinder.mapping:
        _current = MovedModuleFinder.mapping[old_name].new_name
        if new_name == _current:
            return
        raise RuntimeError(
            "Duplicate module alias declaration.\n"
            f"\toriginal: {old_name} -> {_current}\n"
            f"\tconflict: {old_name} -> {new_name}\n"
        )
    MovedModuleFinder.mapping[old_name] = MovedModuleInfo(
        old_name, new_name, msg, logger, version, remove_in
    )



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
# This module was originally developed as part of WaterTAP:
#
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of
# California, through Lawrence Berkeley National Laboratory, Oak Ridge
# National Laboratory, National Renewable Energy Laboratory, and National
# Energy Technology Laboratory (subject to receipt of any required
# approvals from the U.S. Dept. of Energy). All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#
#     Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#
#     Neither the name of the University of California, Lawrence Berkeley
#     National Laboratory, Oak Ridge National Laboratory, National
#     Renewable Energy Laboratory, National Energy Technology Laboratory,
#     U.S. Dept. of Energy nor the names of its contributors may be used
#     to endorse or promote products derived from this software without
#     specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
# IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
# OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# You are under no obligation whatsoever to provide any bug fixes,
# patches, or upgrades to the features, functionality or performance of
# the source code ("Enhancements") to anyone; however, if you choose to
# make your Enhancements available either publicly, or directly to
# Lawrence Berkeley National Laboratory, without imposing a separate
# written license agreement for such Enhancements, then you hereby grant
# the following license: a non-exclusive, royalty-free perpetual license
# to install, use, modify, prepare derivative works, incorporate into
# other computer software, distribute, and sublicense such enhancements or
# derivative works thereof, in binary and source code form.
# ___________________________________________________________________________

"""Minimal Intractable System (MIS) finder

Originally written by Ben Knueven as part of the `WaterTAP project
<https://github.com/watertap-org/watertap>`__.  Copied by DLW 18Feb2024
and edited.

See: http://www.sce.carleton.ca/faculty/chinneck/docs/CPAIOR07InfeasibilityTutorial.pdf

"""

import logging
import pyomo.environ as pyo

from pyomo.core.plugins.transform.add_slack_vars import AddSlackVariables

from pyomo.core.plugins.transform.hierarchy import IsomorphicTransformation

from pyomo.common.modeling import unique_component_name
from pyomo.common.collections import ComponentMap, ComponentSet

from pyomo.opt import WriterFactory

logger = logging.getLogger("pyomo.contrib.iis")
logger.setLevel(logging.INFO)


class _VariableBoundsAsConstraints(IsomorphicTransformation):
    """Replace all variables bounds and domain information with constraints.

    Leaves fixed Vars untouched (for now)
    """

    def _apply_to(self, instance, **kwds):

        bound_constr_block_name = unique_component_name(instance, "_variable_bounds")
        instance.add_component(bound_constr_block_name, pyo.Block())
        bound_constr_block = instance.component(bound_constr_block_name)

        for v in instance.component_data_objects(pyo.Var, descend_into=True):
            if v.fixed:
                continue
            lb, ub = v.bounds
            if lb is None and ub is None:
                continue
            var_name = v.getname(fully_qualified=True)
            if lb is not None:
                con_name = "lb_for_" + var_name
                con = pyo.Constraint(expr=(lb, v, None))
                bound_constr_block.add_component(con_name, con)
            if ub is not None:
                con_name = "ub_for_" + var_name
                con = pyo.Constraint(expr=(None, v, ub))
                bound_constr_block.add_component(con_name, con)

            # now we deactivate the variable bounds / domain
            v.domain = pyo.Reals
            v.setlb(None)
            v.setub(None)


def compute_infeasibility_explanation(
    model, solver, tee=False, tolerance=1e-8, logger=logger
):
    """This function attempts to determine why a given model is infeasible. It deploys
    two main algorithms:

    1. Successfully relaxes the constraints of the problem, and reports to the user
       some sets of constraints and variable bounds, which when relaxed, creates a
       feasible model.
    2. Uses the information collected from (1) to attempt to compute a Minimal
       Infeasible System (MIS), which is a set of constraints and variable bounds
       which appear to be in conflict with each other. It is minimal in the sense
       that removing any single constraint or variable bound would result in a
       feasible subsystem.

    Parameters
    ----------
    model : BlockData
        A pyomo block

    solver: OptSolver | SolverBase | str
        A pyomo solver object or a string for SolverFactory

    tee : bool
        Display intermediate solves conducted

    tolerance : float
        The feasibility tolerance to use when declaring a constraint
        feasible.

    logger: logging.Logger
        A logger for messages. Uses ``pyomo.contrib.mis`` logger by default.

    """
    # Suggested enhancement: It might be useful to return sets of names for each set of relaxed components, as well as the final minimal infeasible system

    # hold the original harmless
    modified_model = model.clone()

    if solver is None:
        raise ValueError("A solver must be supplied")
    elif isinstance(solver, str):
        solver = pyo.SolverFactory(solver)
    else:
        # assume we have a solver
        assert solver.available()

    # first, cache the values we get
    _value_cache = ComponentMap()
    for v in model.component_data_objects(pyo.Var, descend_into=True):
        _value_cache[v] = v.value

    # finding proper reference
    if model.parent_block() is None:
        common_name = ""
    else:
        common_name = model.name + "."

    _modified_model_var_to_original_model_var = ComponentMap()
    _modified_model_value_cache = ComponentMap()

    for v in model.component_data_objects(pyo.Var, descend_into=True):
        modified_model_var = modified_model.find_component(v.name[len(common_name) :])

        _modified_model_var_to_original_model_var[modified_model_var] = v
        _modified_model_value_cache[modified_model_var] = _value_cache[v]
        modified_model_var.set_value(_value_cache[v], skip_validation=True)

    # TODO: For WT / IDAES models, we should probably be more
    #       selective in *what* we elasticize. E.g., it probably
    #       does not make sense to elasticize property calculations
    #       and maybe certain other equality constraints calculating
    #       values. Maybe we shouldn't elasticize *any* equality
    #       constraints.
    #       For example, elasticizing the calculation of mass fraction
    #       makes absolutely no sense and will just be noise for the
    #       modeler to sift through. We could try to sort the constraints
    #       such that we look for those with linear coefficients `1` on
    #       some term and leave those be.
    #       Alternatively, we could apply this tool to a version of the
    #       model that has as many as possible of these constraints
    #       "substituted out".
    # move the variable bounds to the constraints
    _VariableBoundsAsConstraints().apply_to(modified_model)

    AddSlackVariables().apply_to(modified_model)
    slack_block = modified_model._core_add_slack_variables

    for v in slack_block.component_data_objects(pyo.Var):
        v.fix(0)
    # start with variable bounds -- these are the easiest to interpret
    for c in modified_model._variable_bounds.component_data_objects(
        pyo.Constraint, descend_into=True
    ):
        plus = slack_block.component(f"_slack_plus_{c.name}")
        minus = slack_block.component(f"_slack_minus_{c.name}")
        assert not (plus is None and minus is None)
        if plus is not None:
            plus.unfix()
        if minus is not None:
            minus.unfix()

    # TODO: Elasticizing too much at once seems to cause Ipopt trouble.
    #       After an initial sweep, we should just fix one elastic variable
    #       and put everything else on a stack of "constraints to elasticize".
    #       We elasticize one constraint at a time and fix one constraint at a time.
    #       After fixing an elastic variable, we elasticize a single constraint it
    #       appears in and put the remaining constraints on the stack. If the resulting problem
    #       is feasible, we keep going "down the tree". If the resulting problem is
    #       infeasible or cannot be solved, we elasticize a single constraint from
    #       the top of the stack.
    #       The algorithm stops when the stack is empty and the subproblem is infeasible.
    #       Along the way, any time the current problem is infeasible we can check to
    #       see if the current set of constraints in the filter is as a collection of
    #       infeasible constraints -- to terminate early.
    #       However, while more stable, this is much more computationally intensive.
    #       So, we leave the implementation simpler for now and consider this as
    #       a potential extension if this tool sometimes cannot report a good answer.
    # Phase 1 -- build the initial set of constraints, or prove feasibility
    msg = ""
    fixed_slacks = ComponentSet()
    elastic_filter = ComponentSet()

    def _constraint_loop(relaxed_things, msg):
        if msg == "":
            msg += f"Model {model.name} may be infeasible. A feasible solution was found with only the following {relaxed_things} relaxed:\n"
        else:
            msg += f"Another feasible solution was found with only the following {relaxed_things} relaxed:\n"
        while True:

            def _constraint_generator():
                elastic_filter_size_initial = len(elastic_filter)
                for v in slack_block.component_data_objects(pyo.Var):
                    if v.value > tolerance:
                        constr = _get_constraint(modified_model, v)
                        yield constr, v.value
                        v.fix(0)
                        fixed_slacks.add(v)
                        elastic_filter.add(constr)
                if len(elastic_filter) == elastic_filter_size_initial:
                    raise Exception(f"Found model {model.name} to be feasible!")

            msg = _get_results_with_value(_constraint_generator(), msg)
            for var, val in _modified_model_value_cache.items():
                var.set_value(val, skip_validation=True)
            results = solver.solve(modified_model, tee=tee)
            if pyo.check_optimal_termination(results):
                msg += f"Another feasible solution was found with only the following {relaxed_things} relaxed:\n"
            else:
                break
        return msg

    results = solver.solve(modified_model, tee=tee)
    if pyo.check_optimal_termination(results):
        msg = _constraint_loop("variable bounds", msg)

    # next, try relaxing the inequality constraints
    for v in slack_block.component_data_objects(pyo.Var):
        c = _get_constraint(modified_model, v)
        if c.equality:
            # equality constraint
            continue
        if v not in fixed_slacks:
            v.unfix()

    results = solver.solve(modified_model, tee=tee)
    if pyo.check_optimal_termination(results):
        msg = _constraint_loop("inequality constraints and/or variable bounds", msg)

    for v in slack_block.component_data_objects(pyo.Var):
        if v not in fixed_slacks:
            v.unfix()

    results = solver.solve(modified_model, tee=tee)
    if pyo.check_optimal_termination(results):
        msg = _constraint_loop(
            "inequality constraints, equality constraints, and/or variable bounds", msg
        )

    if len(elastic_filter) == 0:
        # load the feasible solution into the original model
        for modified_model_var, v in _modified_model_var_to_original_model_var.items():
            v.set_value(modified_model_var.value, skip_validation=True)
        results = solver.solve(model, tee=tee)
        if pyo.check_optimal_termination(results):
            logger.info(f"A feasible solution was found!")
        else:
            logger.info(
                f"Could not find a feasible solution with violated constraints or bounds. This model is likely unstable"
            )

    # Phase 2 -- deletion filter
    # remove slacks by fixing them to 0
    for v in slack_block.component_data_objects(pyo.Var):
        v.fix(0)
    for o in modified_model.component_data_objects(pyo.Objective, descend_into=True):
        o.deactivate()

    # mark all constraints not in the filter as inactive
    for c in modified_model.component_data_objects(pyo.Constraint):
        if c in elastic_filter:
            continue
        else:
            c.deactivate()

    try:
        results = solver.solve(modified_model, tee=tee)
    except:
        results = None

    if (results is not None) and pyo.check_optimal_termination(results):
        msg += "Could not determine Minimal Intractable System\n"
    else:
        deletion_filter = []
        guards = []
        for constr in elastic_filter:
            constr.deactivate()
            for var, val in _modified_model_value_cache.items():
                var.set_value(val, skip_validation=True)
            math_failure = False
            try:
                results = solver.solve(modified_model, tee=tee)
            except:
                math_failure = True

            if math_failure:
                constr.activate()
                guards.append(constr)
            elif pyo.check_optimal_termination(results):
                constr.activate()
                deletion_filter.append(constr)
            else:  # still infeasible without this constraint
                pass

        msg += "Computed Minimal Intractable System (MIS)!\n"
        msg += "Constraints / bounds in MIS:\n"
        msg = _get_results(deletion_filter, msg)
        msg += "Constraints / bounds in guards for stability:"
        msg = _get_results(guards, msg)

    logger.info(msg)


def _get_results_with_value(constr_value_generator, msg=None):
    # note that "lb_for_" and "ub_for_" are 7 characters long
    if msg is None:
        msg = ""
    for c, value in constr_value_generator:
        c_name = c.name
        if "_variable_bounds" in c_name:
            name = c.local_name
            if "lb" in name:
                msg += f"\tlb of var {name[7:]} by {value}\n"
            elif "ub" in name:
                msg += f"\tub of var {name[7:]} by {value}\n"
            else:
                raise RuntimeError("unrecognized var name")
        else:
            msg += f"\tconstraint: {c_name} by {value}\n"
    return msg


def _get_results(constr_generator, msg=None):
    # note that "lb_for_" and "ub_for_" are 7 characters long
    if msg is None:
        msg = ""
    for c in constr_generator:
        c_name = c.name
        if "_variable_bounds" in c_name:
            name = c.local_name
            if "lb" in name:
                msg += f"\tlb of var {name[7:]}\n"
            elif "ub" in name:
                msg += f"\tub of var {name[7:]}\n"
            else:
                raise RuntimeError("unrecognized var name")
        else:
            msg += f"\tconstraint: {c_name}\n"
    return msg


def _get_constraint(modified_model, v):
    if "_slack_plus_" in v.name:
        constr = modified_model.find_component(v.local_name[len("_slack_plus_") :])
        if constr is None:
            raise RuntimeError(
                f"Bad constraint name {v.local_name[len('_slack_plus_'):]}"
            )
        return constr
    elif "_slack_minus_" in v.name:
        constr = modified_model.find_component(v.local_name[len("_slack_minus_") :])
        if constr is None:
            raise RuntimeError(
                f"Bad constraint name {v.local_name[len('_slack_minus_'):]}"
            )
        return constr
    else:
        raise RuntimeError(f"Bad var name {v.name}")



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
# Unit Tests for complete examples
#

import os
from os.path import abspath, dirname

topdir = dirname(dirname(abspath(__file__))) + os.sep + ".." + os.sep + ".."
currdir = dirname(abspath(__file__)) + os.sep

import pyomo.common.unittest as unittest

from pyomo.environ import *


class TestAmplbook2(unittest.TestCase):
    pass


#
# DISABLED FOR NOW ... Revisit these when the ipconvert utility is stable...
#

# data_dir=topdir+os.sep+"examples"+os.sep+"pyomo"+os.sep+"amplbook2"+os.sep
# files = glob.glob(data_dir+"*.py")
# for file in files:
#    bname=os.path.basename(file)
#    name=bname.split('.')[0]
#    TestAmplbook2.add_commandline_test(cmd='cd '+data_dir+'; '+topdir+os.sep+'scripts/pyomo '+bname+' '+name+'.dat', baseline=data_dir+name+'.log', name=name)

if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import functools
import pickle
import platform
import sys
import types

import pyomo.common.unittest as unittest
from pyomo.common.config import ConfigValue, ConfigList, ConfigDict
from pyomo.common.dependencies import (
    pandas as pd,
    pandas_available,
    numpy as np,
    numpy_available,
)

from pyomo.core.base.util import flatten_tuple
from pyomo.core.base.initializer import (
    Initializer,
    BoundInitializer,
    ConstantInitializer,
    ItemInitializer,
    ScalarCallInitializer,
    IndexedCallInitializer,
    CountedCallInitializer,
    CountedCallGenerator,
    DataFrameInitializer,
    DefaultInitializer,
    ParameterizedInitializer,
    ParameterizedIndexedCallInitializer,
    ParameterizedScalarCallInitializer,
    function_types,
)
from pyomo.environ import ConcreteModel, Var

is_pypy = platform.python_implementation().lower().startswith("pypy")


def _init_scalar(m):
    return 1


def _init_indexed(m, *args):
    i = 1
    for arg in args:
        i *= arg + 1
    return i


class Test_Initializer(unittest.TestCase):
    def test_flattener(self):
        tup = (1, 0, (0, 1), (2, 3))
        self.assertEqual((1, 0, 0, 1, 2, 3), flatten_tuple(tup))
        li = [0]
        self.assertEqual((0,), flatten_tuple(li))
        ex = [(1, 0), [2, 3]]
        self.assertEqual((1, 0, 2, 3), flatten_tuple(ex))

    def test_constant(self):
        a = Initializer(5)
        self.assertIs(type(a), ConstantInitializer)
        self.assertTrue(a.constant())
        self.assertFalse(a.verified)
        self.assertFalse(a.contains_indices())
        with self.assertRaisesRegex(
            RuntimeError,
            "Initializer ConstantInitializer does not contain embedded indices",
        ):
            a.indices()
        self.assertEqual(a(None, 1), 5)

    def test_dict(self):
        a = Initializer({1: 5})
        self.assertIs(type(a), ItemInitializer)
        self.assertFalse(a.constant())
        self.assertFalse(a.verified)
        self.assertTrue(a.contains_indices())
        self.assertEqual(list(a.indices()), [1])
        self.assertEqual(a(None, 1), 5)

    def test_sequence(self):
        a = Initializer([0, 5])
        self.assertIs(type(a), ItemInitializer)
        self.assertFalse(a.constant())
        self.assertFalse(a.verified)
        self.assertTrue(a.contains_indices())
        self.assertEqual(list(a.indices()), [0, 1])
        self.assertEqual(a(None, 1), 5)

        a = Initializer([0, 5], treat_sequences_as_mappings=False)
        self.assertIs(type(a), ConstantInitializer)
        self.assertTrue(a.constant())
        self.assertFalse(a.verified)
        self.assertFalse(a.contains_indices())
        self.assertEqual(a(None, 1), [0, 5])

    def test_function(self):
        def a_init(m):
            return 0

        a = Initializer(a_init)
        self.assertIs(type(a), ScalarCallInitializer)
        self.assertTrue(a.constant())
        self.assertFalse(a.verified)
        self.assertFalse(a.contains_indices())
        self.assertEqual(a(None, 1), 0)

        def x_init(m, i):
            return i + 1

        a = Initializer(x_init)
        self.assertIs(type(a), IndexedCallInitializer)
        self.assertFalse(a.constant())
        self.assertFalse(a.verified)
        self.assertFalse(a.contains_indices())
        self.assertEqual(a(None, 1), 2)

        def x2_init(m):
            return 0

        a = Initializer(x2_init)
        self.assertIs(type(a), ScalarCallInitializer)
        self.assertTrue(a.constant())
        self.assertFalse(a.verified)
        self.assertFalse(a.contains_indices())
        self.assertEqual(a(None, 1), 0)

        def y_init(m, i, j):
            return j * (i + 1)

        a = Initializer(y_init)
        self.assertIs(type(a), IndexedCallInitializer)
        self.assertFalse(a.constant())
        self.assertFalse(a.verified)
        self.assertFalse(a.contains_indices())
        self.assertEqual(a(None, (1, 4)), 8)

    def test_counted_call(self):
        def x_init(m, i):
            return i + 1

        def y_init(m, i, j):
            return j * (i + 1)

        def z_init(m, i, j, k):
            return i * 100 + j * 10 + k

        def bogus(m, i, j):
            return None

        m = ConcreteModel()
        m.x = Var([1, 2, 3])
        a = Initializer(x_init)
        b = CountedCallInitializer(m.x, a)
        self.assertIs(type(b), CountedCallInitializer)
        self.assertFalse(b.constant())
        self.assertFalse(b.verified)
        self.assertFalse(b.contains_indices())
        self.assertFalse(b._scalar)
        self.assertIs(a._fcn, b._fcn)
        c = b(None, 1)
        self.assertIs(type(c), int)
        self.assertEqual(c, 2)

        a = Initializer(bogus)
        b = CountedCallInitializer(m.x, a)
        self.assertIs(type(b), CountedCallInitializer)
        self.assertFalse(b.constant())
        self.assertFalse(b.verified)
        self.assertFalse(b.contains_indices())
        self.assertFalse(b._scalar)
        self.assertIs(a._fcn, b._fcn)
        c = b(None, 1)
        self.assertIs(type(c), CountedCallGenerator)
        with self.assertRaisesRegex(ValueError, 'Counted Var rule returned None'):
            next(c)

        a = Initializer(y_init)
        b = CountedCallInitializer(m.x, a)
        self.assertIs(type(b), CountedCallInitializer)
        self.assertFalse(b.constant())
        self.assertFalse(b.verified)
        self.assertFalse(b.contains_indices())
        self.assertFalse(b._scalar)
        self.assertIs(a._fcn, b._fcn)
        c = b(None, 1)
        self.assertIs(type(c), CountedCallGenerator)
        self.assertEqual(next(c), 2)
        self.assertEqual(next(c), 3)
        self.assertEqual(next(c), 4)

        m.y = Var([(1, 2), (3, 5)])
        a = Initializer(y_init)
        b = CountedCallInitializer(m.y, a)
        self.assertIs(type(b), CountedCallInitializer)
        self.assertFalse(b.constant())
        self.assertFalse(b.verified)
        self.assertFalse(b.contains_indices())
        self.assertFalse(b._scalar)
        self.assertIs(a._fcn, b._fcn)
        c = b(None, (3, 5))
        self.assertIs(type(c), int)
        self.assertEqual(c, 20)

        a = Initializer(z_init)
        b = CountedCallInitializer(m.y, a)
        self.assertIs(type(b), CountedCallInitializer)
        self.assertFalse(b.constant())
        self.assertFalse(b.verified)
        self.assertFalse(b.contains_indices())
        self.assertFalse(b._scalar)
        self.assertIs(a._fcn, b._fcn)
        c = b(None, (3, 5))
        self.assertIs(type(c), CountedCallGenerator)
        self.assertEqual(next(c), 135)
        self.assertEqual(next(c), 235)
        self.assertEqual(next(c), 335)

    def test_method(self):
        class Init:
            def a_init(self, m):
                return 0

            def x_init(self, m, i):
                return i + 1

            def x2_init(self, m):
                return 0

            def y_init(self, m, i, j):
                return j * (i + 1)

        init = Init()

        a = Initializer(init.a_init)
        self.assertIs(type(a), ScalarCallInitializer)
        self.assertTrue(a.constant())
        self.assertFalse(a.verified)
        self.assertFalse(a.contains_indices())
        self.assertEqual(a(None, 1), 0)

        a = Initializer(init.x_init)
        self.assertIs(type(a), IndexedCallInitializer)
        self.assertFalse(a.constant())
        self.assertFalse(a.verified)
        self.assertFalse(a.contains_indices())
        self.assertEqual(a(None, 1), 2)

        a = Initializer(init.x2_init)
        self.assertIs(type(a), ScalarCallInitializer)
        self.assertTrue(a.constant())
        self.assertFalse(a.verified)
        self.assertFalse(a.contains_indices())
        self.assertEqual(a(None, 1), 0)

        a = Initializer(init.y_init)
        self.assertIs(type(a), IndexedCallInitializer)
        self.assertFalse(a.constant())
        self.assertFalse(a.verified)
        self.assertFalse(a.contains_indices())
        self.assertEqual(a(None, (1, 4)), 8)

        m = ConcreteModel()
        m.x = Var([1, 2, 3])
        a = Initializer(init.y_init)
        b = CountedCallInitializer(m.x, a)
        self.assertIs(type(b), CountedCallInitializer)
        self.assertFalse(b.constant())
        self.assertFalse(b.verified)
        self.assertFalse(a.contains_indices())
        self.assertFalse(b._scalar)
        self.assertIs(a._fcn, b._fcn)
        c = b(None, 10)
        self.assertIs(type(c), CountedCallGenerator)
        self.assertEqual(next(c), 20)
        self.assertEqual(next(c), 30)
        self.assertEqual(next(c), 40)

    def test_classmethod(self):
        class Init:
            @classmethod
            def a_init(cls, m):
                return 0

            @classmethod
            def x_init(cls, m, i):
                return i + 1

            @classmethod
            def x2_init(cls, m):
                return 0

            @classmethod
            def y_init(cls, m, i, j):
                return j * (i + 1)

        a = Initializer(Init.a_init)
        self.assertIs(type(a), ScalarCallInitializer)
        self.assertTrue(a.constant())
        self.assertFalse(a.verified)
        self.assertFalse(a.contains_indices())
        self.assertEqual(a(None, 1), 0)

        a = Initializer(Init.x_init)
        self.assertIs(type(a), IndexedCallInitializer)
        self.assertFalse(a.constant())
        self.assertFalse(a.verified)
        self.assertFalse(a.contains_indices())
        self.assertEqual(a(None, 1), 2)

        a = Initializer(Init.x2_init)
        self.assertIs(type(a), ScalarCallInitializer)
        self.assertTrue(a.constant())
        self.assertFalse(a.verified)
        self.assertFalse(a.contains_indices())
        self.assertEqual(a(None, 1), 0)

        a = Initializer(Init.y_init)
        self.assertIs(type(a), IndexedCallInitializer)
        self.assertFalse(a.constant())
        self.assertFalse(a.verified)
        self.assertFalse(a.contains_indices())
        self.assertEqual(a(None, (1, 4)), 8)

        m = ConcreteModel()
        m.x = Var([1, 2, 3])
        a = Initializer(Init.y_init)
        b = CountedCallInitializer(m.x, a)
        self.assertIs(type(b), CountedCallInitializer)
        self.assertFalse(b.constant())
        self.assertFalse(b.verified)
        self.assertFalse(a.contains_indices())
        self.assertFalse(b._scalar)
        self.assertIs(a._fcn, b._fcn)
        c = b(None, 10)
        self.assertIs(type(c), CountedCallGenerator)
        self.assertEqual(next(c), 20)
        self.assertEqual(next(c), 30)
        self.assertEqual(next(c), 40)

    def test_staticmethod(self):
        class Init:
            @staticmethod
            def a_init(m):
                return 0

            @staticmethod
            def x_init(m, i):
                return i + 1

            @staticmethod
            def x2_init(m):
                return 0

            @staticmethod
            def y_init(m, i, j):
                return j * (i + 1)

        a = Initializer(Init.a_init)
        self.assertIs(type(a), ScalarCallInitializer)
        self.assertTrue(a.constant())
        self.assertFalse(a.verified)
        self.assertFalse(a.contains_indices())
        self.assertEqual(a(None, 1), 0)

        a = Initializer(Init.x_init)
        self.assertIs(type(a), IndexedCallInitializer)
        self.assertFalse(a.constant())
        self.assertFalse(a.verified)
        self.assertFalse(a.contains_indices())
        self.assertEqual(a(None, 1), 2)

        a = Initializer(Init.x2_init)
        self.assertIs(type(a), ScalarCallInitializer)
        self.assertTrue(a.constant())
        self.assertFalse(a.verified)
        self.assertFalse(a.contains_indices())
        self.assertEqual(a(None, 1), 0)

        a = Initializer(Init.y_init)
        self.assertIs(type(a), IndexedCallInitializer)
        self.assertFalse(a.constant())
        self.assertFalse(a.verified)
        self.assertFalse(a.contains_indices())
        self.assertEqual(a(None, (1, 4)), 8)

        m = ConcreteModel()
        m.x = Var([1, 2, 3])
        a = Initializer(Init.y_init)
        b = CountedCallInitializer(m.x, a)
        self.assertIs(type(b), CountedCallInitializer)
        self.assertFalse(b.constant())
        self.assertFalse(b.verified)
        self.assertFalse(b.contains_indices())
        self.assertFalse(b._scalar)
        self.assertIs(a._fcn, b._fcn)
        c = b(None, 10)
        self.assertIs(type(c), CountedCallGenerator)
        self.assertEqual(next(c), 20)
        self.assertEqual(next(c), 30)
        self.assertEqual(next(c), 40)

    def test_generator_fcn(self):
        def a_init(m):
            yield 0
            yield 3

        with self.assertRaisesRegex(ValueError, "Generator functions are not allowed"):
            a = Initializer(a_init)

        a = Initializer(a_init, allow_generators=True)
        self.assertIs(type(a), ScalarCallInitializer)
        self.assertFalse(a.constant())
        self.assertFalse(a.verified)
        self.assertEqual(list(a(None, 1)), [0, 3])

        def x_init(m, i):
            yield i
            yield i + 1

        a = Initializer(x_init, allow_generators=True)
        self.assertIs(type(a), IndexedCallInitializer)
        self.assertFalse(a.constant())
        self.assertFalse(a.verified)
        self.assertEqual(list(a(None, 1)), [1, 2])

        def y_init(m, i, j):
            yield j
            yield i + 1

        a = Initializer(y_init, allow_generators=True)
        self.assertIs(type(a), IndexedCallInitializer)
        self.assertFalse(a.constant())
        self.assertFalse(a.verified)
        self.assertEqual(list(a(None, (1, 4))), [4, 2])

    def test_generator_method(self):
        class Init:
            def a_init(self, m):
                yield 0
                yield 3

            def x_init(self, m, i):
                yield i
                yield i + 1

            def y_init(self, m, i, j):
                yield j
                yield i + 1

        init = Init()

        with self.assertRaisesRegex(ValueError, "Generator functions are not allowed"):
            a = Initializer(init.a_init)

        a = Initializer(init.a_init, allow_generators=True)
        self.assertIs(type(a), ScalarCallInitializer)
        self.assertFalse(a.constant())
        self.assertFalse(a.verified)
        self.assertEqual(list(a(None, 1)), [0, 3])

        a = Initializer(init.x_init, allow_generators=True)
        self.assertIs(type(a), IndexedCallInitializer)
        self.assertFalse(a.constant())
        self.assertFalse(a.verified)
        self.assertEqual(list(a(None, 1)), [1, 2])

        a = Initializer(init.y_init, allow_generators=True)
        self.assertIs(type(a), IndexedCallInitializer)
        self.assertFalse(a.constant())
        self.assertFalse(a.verified)
        self.assertEqual(list(a(None, (1, 4))), [4, 2])

    def test_generators(self):
        with self.assertRaisesRegex(ValueError, "Generators are not allowed"):
            a = Initializer(iter([0, 3]))

        a = Initializer(iter([0, 3]), allow_generators=True)
        self.assertIs(type(a), ConstantInitializer)
        self.assertTrue(a.constant())
        self.assertFalse(a.verified)
        self.assertEqual(list(a(None, 1)), [0, 3])

        def x_init():
            yield 0
            yield 3

        with self.assertRaisesRegex(ValueError, "Generators are not allowed"):
            a = Initializer(x_init())

        a = Initializer(x_init(), allow_generators=True)
        self.assertIs(type(a), ConstantInitializer)
        self.assertTrue(a.constant())
        self.assertFalse(a.verified)
        self.assertEqual(list(a(None, 1)), [0, 3])

    def test_functor(self):
        class InitScalar:
            def __init__(self, val):
                self.val = val

            def __call__(self, m):
                return self.val

        a = Initializer(InitScalar(10))
        self.assertIs(type(a), ScalarCallInitializer)
        self.assertTrue(a.constant())
        self.assertFalse(a.verified)
        self.assertEqual(a(None, None), 10)

        class InitIndexed:
            def __init__(self, val):
                self.val = val

            def __call__(self, m, i):
                return self.val + i

        a = Initializer(InitIndexed(10))
        self.assertIs(type(a), IndexedCallInitializer)
        self.assertFalse(a.constant())
        self.assertFalse(a.verified)
        self.assertEqual(a(None, 5), 15)

    def test_derived_function(self):
        def _scalar(m):
            return 10

        dynf = types.FunctionType(_scalar.__code__, {})

        a = Initializer(dynf)
        self.assertIs(type(a), ScalarCallInitializer)
        self.assertTrue(a.constant())
        self.assertFalse(a.verified)
        self.assertEqual(a(None, None), 10)

        def _indexed(m, i):
            return 10 + i

        dynf = types.FunctionType(_indexed.__code__, {})

        a = Initializer(dynf)
        self.assertIs(type(a), IndexedCallInitializer)
        self.assertFalse(a.constant())
        self.assertFalse(a.verified)
        self.assertEqual(a(None, 5), 15)

    def test_function(self):
        def _scalar(m):
            return 10

        a = Initializer(_scalar)
        self.assertIs(type(a), ScalarCallInitializer)
        self.assertTrue(a.constant())
        self.assertFalse(a.verified)
        self.assertEqual(a(None, None), 10)

        def _indexed(m, i):
            return 10 + i

        a = Initializer(_indexed)
        self.assertIs(type(a), IndexedCallInitializer)
        self.assertFalse(a.constant())
        self.assertFalse(a.verified)
        self.assertEqual(a(None, 5), 15)

        try:
            original_fcn_types = set(function_types)
            function_types.clear()
            self.assertEqual(len(function_types), 0)

            a = Initializer(_scalar)
            self.assertIs(type(a), ScalarCallInitializer)
            self.assertTrue(a.constant())
            self.assertFalse(a.verified)
            self.assertEqual(a(None, None), 10)
            self.assertEqual(len(function_types), 1)
        finally:
            function_types.clear()
            function_types.update(original_fcn_types)

        try:
            original_fcn_types = set(function_types)
            function_types.clear()
            self.assertEqual(len(function_types), 0)

            a = Initializer(_indexed)
            self.assertIs(type(a), IndexedCallInitializer)
            self.assertFalse(a.constant())
            self.assertFalse(a.verified)
            self.assertEqual(a(None, 5), 15)
        finally:
            function_types.clear()
            function_types.update(original_fcn_types)

    def test_no_argspec(self):
        a = Initializer(getattr)
        self.assertIs(type(a), IndexedCallInitializer)
        self.assertFalse(a.constant())
        self.assertFalse(a.verified)
        self.assertFalse(a.contains_indices())
        self.assertEqual(a(None, '__class__'), type(None))

        basetwo = functools.partial(int)
        a = Initializer(basetwo)
        self.assertIs(type(a), IndexedCallInitializer)
        self.assertFalse(a.constant())
        self.assertFalse(a.verified)
        self.assertFalse(a.contains_indices())
        self.assertEqual(a('111', 2), 7)

        # Special case: getfullargspec fails for int under CPython and
        # PyPy<7.3.14, so we assume it is an IndexedCallInitializer.
        basetwo = functools.partial(int, '101', base=2)
        a = Initializer(basetwo)
        if is_pypy and sys.pypy_version_info[:3] >= (7, 3, 14):
            # PyPy behavior diverged from CPython in 7.3.14.  Arguably
            # this is "more correct", so we will allow the difference to
            # persist through Pyomo's Initializer handling (and not
            # special case it there)
            self.assertIs(type(a), ScalarCallInitializer)
            self.assertTrue(a.constant())
        else:
            self.assertIs(type(a), IndexedCallInitializer)
            self.assertFalse(a.constant())
        self.assertFalse(a.verified)
        self.assertFalse(a.contains_indices())
        # but this is not callable, as int won't accept the 'model'
        # self.assertEqual(a(None, None), 5)

    def test_partial(self):
        def fcn(k, m, i, j):
            return i * 100 + j * 10 + k

        part = functools.partial(fcn, 2)
        a = Initializer(part)
        self.assertIs(type(a), IndexedCallInitializer)
        self.assertFalse(a.constant())
        self.assertFalse(a.verified)
        self.assertFalse(a.contains_indices())
        self.assertEqual(a(None, (5, 7)), 572)

        def fcn(k, i, j, m):
            return i * 100 + j * 10 + k

        part = functools.partial(fcn, 2, 5, 7)
        a = Initializer(part)
        self.assertIs(type(a), ScalarCallInitializer)
        self.assertTrue(a.constant())
        self.assertFalse(a.verified)
        self.assertFalse(a.contains_indices())
        self.assertEqual(a(None, None), 572)

        def fcn(m, k, i, j):
            return i * 100 + j * 10 + k

        part = functools.partial(fcn, i=2, j=5, k=7)
        a = Initializer(part)
        self.assertIs(type(a), ScalarCallInitializer)
        self.assertTrue(a.constant())
        self.assertFalse(a.verified)
        self.assertFalse(a.contains_indices())
        self.assertEqual(a(None, None), 257)

    @unittest.skipUnless(pandas_available, "Pandas is not installed")
    def test_dataframe(self):
        d = {'col1': [1, 2, 4]}
        df = pd.DataFrame(data=d)
        a = Initializer(df)
        self.assertIs(type(a), DataFrameInitializer)
        self.assertFalse(a.constant())
        self.assertFalse(a.verified)
        self.assertTrue(a.contains_indices())
        self.assertEqual(list(a.indices()), [0, 1, 2])
        self.assertEqual(a(None, 0), 1)
        self.assertEqual(a(None, 1), 2)
        self.assertEqual(a(None, 2), 4)

        d = {'col1': [1, 2, 4], 'col2': [10, 20, 40]}
        df = pd.DataFrame(data=d)
        a = Initializer(df)
        self.assertIs(type(a), DataFrameInitializer)
        self.assertFalse(a.constant())
        self.assertFalse(a.verified)
        self.assertTrue(a.contains_indices())
        self.assertEqual(
            list(a.indices()),
            [
                (0, 'col1'),
                (0, 'col2'),
                (1, 'col1'),
                (1, 'col2'),
                (2, 'col1'),
                (2, 'col2'),
            ],
        )
        self.assertEqual(a(None, (0, 'col1')), 1)
        self.assertEqual(a(None, (1, 'col2')), 20)
        self.assertEqual(a(None, (2, 'col2')), 40)

        a = DataFrameInitializer(df, 'col2')
        self.assertIs(type(a), DataFrameInitializer)
        self.assertFalse(a.constant())
        self.assertFalse(a.verified)
        self.assertTrue(a.contains_indices())
        self.assertEqual(list(a.indices()), [0, 1, 2])
        self.assertEqual(a(None, 0), 10)
        self.assertEqual(a(None, 1), 20)
        self.assertEqual(a(None, 2), 40)

        df = pd.DataFrame([10, 20, 30, 40], index=[[0, 0, 1, 1], [0, 1, 0, 1]])
        a = Initializer(df)
        self.assertIs(type(a), DataFrameInitializer)
        self.assertFalse(a.constant())
        self.assertFalse(a.verified)
        self.assertTrue(a.contains_indices())
        self.assertEqual(list(a.indices()), [(0, 0), (0, 1), (1, 0), (1, 1)])
        self.assertEqual(a(None, (0, 0)), 10)
        self.assertEqual(a(None, (0, 1)), 20)
        self.assertEqual(a(None, (1, 0)), 30)
        self.assertEqual(a(None, (1, 1)), 40)

    @unittest.skipUnless(pandas_available, "Pandas is not installed")
    def test_series(self):
        d = pd.Series({0: 1, 1: 2, 2: 4})
        a = Initializer(d)
        self.assertIs(type(a), ItemInitializer)
        self.assertFalse(a.constant())
        self.assertFalse(a.verified)
        self.assertTrue(a.contains_indices())
        self.assertEqual(list(a.indices()), [0, 1, 2])
        self.assertEqual(a(None, 0), 1)
        self.assertEqual(a(None, 1), 2)
        self.assertEqual(a(None, 2), 4)

    @unittest.skipUnless(numpy_available, "Numpy is not installed")
    def test_ndarray(self):
        d = np.array([1, 2, 4])
        a = Initializer(d)
        self.assertIs(type(a), ItemInitializer)
        self.assertFalse(a.constant())
        self.assertFalse(a.verified)
        self.assertTrue(a.contains_indices())
        self.assertEqual(list(a.indices()), [0, 1, 2])
        self.assertEqual(a(None, 0), 1)
        self.assertEqual(a(None, 1), 2)
        self.assertEqual(a(None, 2), 4)

        # TODO: How should we handle ndarray matrices?
        # d = np.array([[1,2],[4,6]])
        # a = Initializer(d)
        # self.assertIs(type(a), ItemInitializer)
        # self.assertFalse(a.constant())
        # self.assertFalse(a.verified)
        # self.assertTrue(a.contains_indices())
        # self.assertEqual(list(a.indices()), [0,1,2])
        # self.assertEqual(a(None, 0), 1)
        # self.assertEqual(a(None, 1), 2)
        # self.assertEqual(a(None, 2), 4)

    def test_str(self):
        a = Initializer("a string")
        self.assertIs(type(a), ConstantInitializer)
        self.assertTrue(a.constant())
        self.assertFalse(a.verified)
        self.assertFalse(a.contains_indices())
        self.assertEqual(a(None, None), "a string")

    @unittest.skipUnless(pandas_available, "Pandas is not installed")
    def test_initializer_initializer(self):
        d = {'col1': [1, 2, 4], 'col2': [10, 20, 40]}
        df = pd.DataFrame(data=d)
        a = Initializer(DataFrameInitializer(df, 'col2'))
        self.assertIs(type(a), DataFrameInitializer)
        self.assertFalse(a.constant())
        self.assertFalse(a.verified)
        self.assertTrue(a.contains_indices())
        self.assertEqual(list(a.indices()), [0, 1, 2])
        self.assertEqual(a(None, 0), 10)
        self.assertEqual(a(None, 1), 20)
        self.assertEqual(a(None, 2), 40)

    def test_pickle(self):
        a = Initializer(5)
        a.verified = True
        b = pickle.loads(pickle.dumps(a))
        self.assertIsNot(a, b)
        self.assertEqual(a.val, b.val)
        self.assertEqual(a.verified, b.verified)

        a = Initializer({1: 5})
        b = pickle.loads(pickle.dumps(a))
        self.assertIsNot(a, b)
        self.assertEqual(a._dict, b._dict)
        self.assertIsNot(a._dict, b._dict)
        self.assertEqual(a.verified, b.verified)

        a = Initializer(_init_scalar)
        b = pickle.loads(pickle.dumps(a))
        self.assertIsNot(a, b)
        self.assertIs(a._fcn, b._fcn)
        self.assertEqual(a.verified, b.verified)
        self.assertEqual(a(None, None), 1)
        self.assertEqual(b(None, None), 1)

        a = Initializer(_init_indexed)
        b = pickle.loads(pickle.dumps(a))
        self.assertIsNot(a, b)
        self.assertIs(a._fcn, b._fcn)
        self.assertEqual(a.verified, b.verified)
        self.assertEqual(a(None, 1), 2)
        self.assertEqual(b(None, 2), 3)

    def test_default_initializer(self):
        a = Initializer({1: 5})
        d = DefaultInitializer(a, None, KeyError)
        self.assertFalse(d.constant())
        self.assertTrue(d.contains_indices())
        self.assertEqual(list(d.indices()), [1])
        self.assertEqual(d(None, 1), 5)
        self.assertIsNone(d(None, 2))

        def rule(m, i):
            if i == 0:
                return 10
            elif i == 1:
                raise KeyError("key")
            elif i == 2:
                raise TypeError("type")
            else:
                raise RuntimeError("runtime")

        a = Initializer(rule)
        d = DefaultInitializer(a, 100, (KeyError, RuntimeError))
        self.assertFalse(d.constant())
        self.assertFalse(d.contains_indices())
        self.assertEqual(d(None, 0), 10)
        self.assertEqual(d(None, 1), 100)
        with self.assertRaisesRegex(TypeError, 'type'):
            d(None, 2)
        self.assertEqual(d(None, 3), 100)

    def test_config_integration(self):
        c = ConfigList()
        c.add(1)
        c.add(3)
        c.add(5)
        a = Initializer(c)
        self.assertIs(type(a), ItemInitializer)
        self.assertTrue(a.contains_indices())
        self.assertEqual(list(a.indices()), [0, 1, 2])
        self.assertEqual(a(None, 0), 1)
        self.assertEqual(a(None, 1), 3)
        self.assertEqual(a(None, 2), 5)

        c = ConfigDict()
        c.declare('opt_1', ConfigValue(default=1))
        c.declare('opt_3', ConfigValue(default=3))
        c.declare('opt_5', ConfigValue(default=5))
        a = Initializer(c)
        self.assertIs(type(a), ItemInitializer)
        self.assertTrue(a.contains_indices())
        self.assertEqual(list(a.indices()), ['opt_1', 'opt_3', 'opt_5'])
        self.assertEqual(a(None, 'opt_1'), 1)
        self.assertEqual(a(None, 'opt_3'), 3)
        self.assertEqual(a(None, 'opt_5'), 5)

    def _bound_function1(self, m, i):
        return m, i

    def _bound_function2(self, m, i, j):
        return m, i, j

    def test_additional_args(self):
        def a_init(m):
            yield 0
            yield 3

        with self.assertRaisesRegex(
            ValueError,
            "Generator functions are not allowed when passing additional args",
        ):
            a = Initializer(a_init, additional_args=1)

        a = Initializer(self._bound_function1, additional_args=1)
        self.assertIs(type(a), ParameterizedScalarCallInitializer)
        self.assertEqual(a('m', None, 5), ('m', 5))

        a = Initializer(self._bound_function2, additional_args=1)
        self.assertIs(type(a), ParameterizedIndexedCallInitializer)
        self.assertEqual(a('m', 1, 5), ('m', 5, 1))

        class Functor:
            def __init__(self, i):
                self.i = i

            def __call__(self, m, i):
                return m, i * self.i

        a = Initializer(Functor(10), additional_args=1)
        self.assertIs(type(a), ParameterizedScalarCallInitializer)
        self.assertEqual(a('m', None, 5), ('m', 50))

        a_init = {1: lambda m, i: ('m', i), 2: lambda m, i: ('m', 2 * i)}
        a = Initializer(a_init, additional_args=1)
        self.assertIs(type(a), ParameterizedInitializer)
        self.assertFalse(a.constant())
        self.assertTrue(a.contains_indices())
        self.assertEqual(list(a.indices()), [1, 2])
        self.assertEqual(a('m', 1, 5), ('m', 5))
        self.assertEqual(a('m', 2, 5), ('m', 10))

    def test_bound_initializer(self):
        m = ConcreteModel()
        m.x = Var([0, 1, 2])
        m.y = Var()

        b = BoundInitializer(None, m.x)
        self.assertIsNone(b)

        b = BoundInitializer((0, 1), m.x)
        self.assertIs(type(b), BoundInitializer)
        self.assertTrue(b.constant())
        self.assertFalse(b.verified)
        self.assertFalse(b.contains_indices())
        self.assertEqual(b(None, 1), (0, 1))

        b = BoundInitializer([(0, 1)], m.x)
        self.assertIs(type(b), BoundInitializer)
        self.assertFalse(b.constant())
        self.assertFalse(b.verified)
        self.assertTrue(b.contains_indices())
        self.assertTrue(list(b.indices()), [0])
        self.assertEqual(b(None, 0), (0, 1))

        init = {1: (2, 3), 4: (5, 6)}
        b = BoundInitializer(init, m.x)
        self.assertIs(type(b), BoundInitializer)
        self.assertFalse(b.constant())
        self.assertFalse(b.verified)
        self.assertTrue(b.contains_indices())
        self.assertEqual(list(b.indices()), [1, 4])
        self.assertEqual(b(None, 1), (2, 3))
        self.assertEqual(b(None, 4), (5, 6))

        b = BoundInitializer((0, 1), m.y)
        self.assertEqual(b(None, None), (0, 1))

        b = BoundInitializer(5, m.y)
        self.assertEqual(b(None, None), (5, 5))



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
# Development of this module was conducted as part of the Institute for
# the Design of Advanced Energy Systems (IDAES) with support through the
# Simulation-Based Engineering, Crosscutting Research Program within the
# U.S. Department of Energy's Office of Fossil Energy and Carbon Management.
# ____________________________________________________________________________________

"""
Common utilities for Trust Region Framework
"""

import logging

logger = logging.getLogger('pyomo.contrib.trustregion')


def minIgnoreNone(a, b):
    """
    Return the min of two numbers, ignoring None
    """
    if a is None:
        return b
    if b is None:
        return a
    if a < b:
        return a
    return b


def maxIgnoreNone(a, b):
    """
    Return the max of two numbers, ignoring None
    """
    if a is None:
        return b
    if b is None:
        return a
    if a < b:
        return b
    return a


class IterationRecord:
    """
    Record relevant information at each individual iteration
    """

    def __init__(
        self,
        iteration,
        feasibility=None,
        objectiveValue=None,
        trustRadius=None,
        stepNorm=None,
    ):
        self.iteration = iteration
        self.fStep, self.thetaStep, self.rejected = [False] * 3
        if feasibility is not None:
            self.feasibility = feasibility
        if objectiveValue is not None:
            self.objectiveValue = objectiveValue
        if trustRadius is not None:
            self.trustRadius = trustRadius
        if stepNorm is not None:
            self.stepNorm = stepNorm

    def detailLogger(self):
        """
        Print information about the iteration to the log.
        """
        logger.info("****** Iteration %d ******" % self.iteration)
        logger.info("trustRadius = %s" % self.trustRadius)
        logger.info("feasibility = %s" % self.feasibility)
        logger.info("objectiveValue = %s" % self.objectiveValue)
        logger.info("stepNorm = %s" % self.stepNorm)
        if self.fStep:
            logger.info("INFO: f-type step")
        if self.thetaStep:
            logger.info("INFO: theta-type step")
        if self.rejected:
            logger.info("INFO: step rejected")

    def verboseLogger(self):
        """
        Print information about the iteration to the console
        """
        print("****** Iteration %d ******" % self.iteration)
        print("objectiveValue = %s" % self.objectiveValue)
        print("feasibility = %s" % self.feasibility)
        print("trustRadius = %s" % self.trustRadius)
        print("stepNorm = %s" % self.stepNorm)
        if self.fStep:
            print("INFO: f-type step")
        if self.thetaStep:
            print("INFO: theta-type step")
        if self.rejected:
            print("INFO: step rejected")
        print(25 * '*')


class IterationLogger:
    """
    Log (and print) information for all iterations
    """

    def __init__(self):
        self.iterations = []

    def newIteration(
        self, iteration, feasibility, objectiveValue, trustRadius, stepNorm
    ):
        """
        Add a new iteration to the list of iterations
        """
        self.iterrecord = IterationRecord(
            iteration,
            feasibility=feasibility,
            objectiveValue=objectiveValue,
            trustRadius=trustRadius,
            stepNorm=stepNorm,
        )
        self.iterations.append(self.iterrecord)

    def updateIteration(
        self, feasibility=None, objectiveValue=None, trustRadius=None, stepNorm=None
    ):
        """
        Update values in current record
        """
        if feasibility is not None:
            self.iterrecord.feasibility = feasibility
        if objectiveValue is not None:
            self.iterrecord.objectiveValue = objectiveValue
        if trustRadius is not None:
            self.iterrecord.trustRadius = trustRadius
        if stepNorm is not None:
            self.iterrecord.stepNorm = stepNorm

    def logIteration(self):
        """
        Log detailed information about the iteration to the log
        """
        self.iterrecord.detailLogger()

    def printIteration(self):
        """
        Print information to the screen
        """
        self.iterrecord.verboseLogger()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.contrib.mpc.data.dynamic_data_base import _is_iterable
from pyomo.contrib.mpc.data.find_nearest_index import (
    find_nearest_index,
    find_nearest_interval_index,
)


def _raise_invalid_cuid(cuid, model):
    raise RuntimeError("Cannot find a component %s on block %s" % (cuid, model))


def load_data_from_scalar(data, model, time):
    """A function to load ScalarData into a model

    Arguments
    ---------
    data: ~scalar_data.ScalarData
    model: BlockData
    time: Iterable

    """
    data = data.get_data()
    t_iter = time if _is_iterable(time) else (time,)
    for cuid, val in data.items():
        var = model.find_component(cuid)
        if var is None:
            _raise_invalid_cuid(cuid, model)
        # TODO: Time points should probably use find_nearest_index
        # This will have to happen in the calling function, as data
        # doesn't have a list of time points to check.
        if var.is_indexed():
            for t in t_iter:
                var[t].set_value(val)
        else:
            var.set_value(val)


def load_data_from_series(data, model, time, tolerance=0.0):
    """A function to load TimeSeriesData into a model

    Arguments
    ---------
    data: TimeSeriesData
    model: BlockData
    time: Iterable

    """
    time_list = list(time)
    time_indices = [
        find_nearest_index(time_list, t, tolerance=tolerance)
        for t in data.get_time_points()
    ]
    for idx, t in zip(time_indices, data.get_time_points()):
        if idx is None:
            raise RuntimeError("Time point %s not found time set" % t)
    if len(time_list) != len(data.get_time_points()):
        raise RuntimeError(
            "TimeSeriesData object and model must have same number"
            " of time points to load data from series"
        )
    data = data.get_data()
    for cuid, vals in data.items():
        var = model.find_component(cuid)
        if var is None:
            _raise_invalid_cuid(cuid, model)
        for idx, val in zip(time_indices, vals):
            t = time_list[idx]
            var[t].set_value(val)


def load_data_from_interval(
    data,
    model,
    time,
    tolerance=0.0,
    prefer_left=True,
    exclude_left_endpoint=True,
    exclude_right_endpoint=False,
):
    """A function to load IntervalData into a model

    Loads values into specified variables at time points that are
    within the intervals specified. If a time point is on the boundary
    of two intervals, the default is to use the interval on the left.
    Often, intervals should be treated as half-open, i.e. one of the
    left or right endpoints should be excluded. This can be enforced
    with the corresponding optional arguments.

    Arguments
    ---------
    data: IntervalData
    model: BlockData
    time: Iterable
    tolerance: Float
    prefer_left: Bool
    exclude_left_endpoint: Bool
    exclude_right_endpoint: Bool

    """
    if prefer_left and exclude_right_endpoint and not exclude_left_endpoint:
        raise RuntimeError(
            "Cannot use prefer_left=True with exclude_left_endpoint=False"
            " and exclude_right_endpoint=True."
        )
    elif not prefer_left and exclude_left_endpoint and not exclude_right_endpoint:
        raise RuntimeError(
            "Cannot use prefer_left=False with exclude_left_endpoint=True"
            " and exclude_right_endpoint=False."
        )
    intervals = data.get_intervals()
    left_endpoints = [t for t, _ in intervals]
    right_endpoints = [t for _, t in intervals]
    # NOTE: O(len(time)*log(len(intervals)))
    idx_list = [
        find_nearest_interval_index(
            intervals, t, tolerance=tolerance, prefer_left=prefer_left
        )
        for t in time
    ]
    left_endpoint_indices = [
        # index of interval which t is the left endpoint of
        find_nearest_index(left_endpoints, t, tolerance=tolerance)
        for t in time
    ]
    right_endpoint_indices = [
        # index of interval which t is the right endpoint of
        find_nearest_index(right_endpoints, t, tolerance=tolerance)
        for t in time
    ]

    # Post-process indices to exclude endpoints
    for i, t in enumerate(time):
        if (
            exclude_left_endpoint
            and left_endpoint_indices[i] is not None
            and right_endpoint_indices[i] is None
        ):
            # If t is a left endpoint but not a right endpoint,
            # do not load a value at t.
            idx_list[i] = None
        elif (
            exclude_right_endpoint
            and right_endpoint_indices[i] is not None
            and left_endpoint_indices[i] is None
        ):
            # If t is a right endpoint but not a left endpoint,
            # do not load a value at t.
            idx_list[i] = None
        elif (
            exclude_left_endpoint
            and exclude_right_endpoint
            and right_endpoint_indices[i] is not None
            and left_endpoint_indices[i] is not None
        ):
            # t is both a left endpoint and a right endpoint
            idx_list[i] = None

    data = data.get_data()
    for cuid, vals in data.items():
        var = model.find_component(cuid)
        if var is None:
            _raise_invalid_cuid(cuid, model)
        for i, t in zip(idx_list, time):
            if i is None:
                # t could not be found in an interval. This is fine.
                # We don't necessarily require that the interval data
                # cover the entire time set.
                continue
            else:
                var[t].set_value(vals[i])



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.contrib.piecewise import PiecewiseLinearFunction, Triangulation
from pyomo.environ import ConcreteModel, Constraint, log, Objective, Var

default_simplices = [
    [(0, 1), (0, 4), (3, 4)],
    [(0, 1), (3, 4), (3, 1)],
    [(3, 4), (3, 7), (0, 7)],
    [(0, 7), (0, 4), (3, 4)],
]


def make_log_x_model(simplices=default_simplices):
    m = ConcreteModel()
    m.x = Var(bounds=(1, 10))
    m.pw_log = PiecewiseLinearFunction(points=[1, 3, 6, 10], function=log)

    # Here are the linear functions, for safe keeping.
    def f1(x):
        return (log(3) / 2) * x - log(3) / 2

    m.f1 = f1

    def f2(x):
        return (log(2) / 3) * x + log(3 / 2)

    m.f2 = f2

    def f3(x):
        return (log(5 / 3) / 4) * x + log(6 / ((5 / 3) ** (3 / 2)))

    m.f3 = f3

    m.log_expr = m.pw_log(m.x)
    m.obj = Objective(expr=m.log_expr)

    m.x1 = Var(bounds=(0, 3))
    m.x2 = Var(bounds=(1, 7))

    ## apprximates paraboloid x1**2 + x2**2
    def g1(x1, x2):
        return 3 * x1 + 5 * x2 - 4

    m.g1 = g1

    def g2(x1, x2):
        return 3 * x1 + 11 * x2 - 28

    m.g2 = g2

    m.pw_paraboloid = PiecewiseLinearFunction(
        simplices=simplices,
        linear_functions=[g1, g1, g2, g2],
        triangulation=Triangulation.AssumeValid,
    )
    m.paraboloid_expr = m.pw_paraboloid(m.x1, m.x2)

    def c_rule(m, i):
        if i == 0:
            return m.x >= m.paraboloid_expr
        else:
            return (1, m.x1, 2)

    m.indexed_c = Constraint([0, 1], rule=c_rule)

    return m



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
# Unit Tests for util/misc
#

import os
from os.path import abspath, dirname

pyomodir = dirname(abspath(__file__)) + "/../.."
currdir = dirname(abspath(__file__)) + os.sep

import pyomo.common.unittest as unittest

from pyomo.common.tempfiles import TempfileManager
import pyomo.opt
import pyomo.opt.plugins.sol
from pyomo.opt.base.solvers import UnknownSolver

old_tempdir = TempfileManager.tempdir


class MockWriter(pyomo.opt.AbstractProblemWriter):
    def __init__(self, name=None):
        pyomo.opt.AbstractProblemWriter.__init__(self, name)


class MockReader(pyomo.opt.AbstractResultsReader):
    def __init__(self, name=None):
        pyomo.opt.AbstractResultsReader.__init__(self, name)


class MockSolver(pyomo.opt.OptSolver):
    def __init__(self, **kwds):
        kwds['type'] = 'stest_type'
        kwds['doc'] = 'MockSolver Documentation'
        pyomo.opt.OptSolver.__init__(self, **kwds)


class Test(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        import pyomo.environ

    def run(self, result=None):
        unittest.TestCase.run(self, result)

    def setUp(self):
        pyomo.opt.WriterFactory.register('wtest')(MockWriter)
        pyomo.opt.ReaderFactory.register('rtest')(MockReader)
        pyomo.opt.SolverFactory.register('stest')(MockSolver)
        TempfileManager.tempdir = currdir

    def tearDown(self):
        TempfileManager.clear_tempfiles()
        TempfileManager.tempdir = old_tempdir
        pyomo.opt.WriterFactory.unregister('wtest')
        pyomo.opt.ReaderFactory.unregister('rtest')
        pyomo.opt.SolverFactory.unregister('stest')

    def test_solver_factory(self):
        # """
        # Testing the pyomo.opt solver factory
        # """
        ans = sorted(list(pyomo.opt.SolverFactory))
        # self.assertEqual(len(ans),8)
        self.assertTrue(set(['stest']) <= set(ans))

    def test_solver_instance(self):
        # """
        # Testing that we get a specific solver instance
        # """
        ans = pyomo.opt.SolverFactory("none")
        self.assertTrue(isinstance(ans, UnknownSolver))
        ans = pyomo.opt.SolverFactory("stest")
        self.assertEqual(type(ans), MockSolver)
        ans = pyomo.opt.SolverFactory("stest", name="mymock")
        self.assertEqual(type(ans), MockSolver)
        self.assertEqual(ans.name, "mymock")

    def test_writer_factory(self):
        # """
        # Testing the pyomo.opt writer factory
        # """
        factory = pyomo.opt.WriterFactory
        self.assertTrue(set(['wtest']) <= set(factory))

    def test_writer_instance(self):
        # """
        # Testing that we get a specific writer instance
        #
        # Note: this simply provides code coverage right now, but
        # later it should be adapted to generate a specific writer.
        # """
        ans = pyomo.opt.WriterFactory("none")
        self.assertEqual(ans, None)
        ans = pyomo.opt.WriterFactory("wtest")
        self.assertNotEqual(ans, None)

    def test_reader_factory(self):
        # """
        # Testing the pyomo.opt reader factory
        # """
        ans = pyomo.opt.ReaderFactory
        self.assertTrue(set(ans) >= set(["rtest", "sol", "yaml", "json"]))

    def test_reader_instance(self):
        # """
        # Testing that we get a specific reader instance
        # """
        ans = pyomo.opt.ReaderFactory("none")
        self.assertEqual(ans, None)
        ans = pyomo.opt.ReaderFactory("sol")
        self.assertEqual(type(ans), pyomo.opt.plugins.sol.ResultsReader_sol)
        # ans = pyomo.opt.ReaderFactory("osrl", "myreader")
        # self.assertEqual(type(ans), pyomo.opt.reader.OS.ResultsReader_osrl)
        # self.assertEqual(ans.name, "myreader")


if __name__ == "__main__":
    unittest.main()
